import itertools
import socket as socket_module
import sys
import time
import random
import threading
from functools import partial
from typing import Optional, List, Dict, Tuple, Sequence

from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.watchdog_wrapper import PathInfo, IFSCallback, IFSWatch
import os

log = get_logger(__name__)


class _RemoteFSWatch(object):
    """
    Holds together the information related to watching a filesystem path
    along with a way to stop watching it.
    """

    def __init__(
        self,
        remote_fs_observer: "RemoteFSObserver",
        on_change_id: str,
        on_change: IFSCallback,
        call_args=(),
    ):
        self.remote_fs_observer = remote_fs_observer
        self.on_change_id = on_change_id
        self.on_change = on_change
        self.call_args = call_args
        self.acknowledged = threading.Event()

    def stop_tracking(self):
        # Note that to stop tracking we actually send a message back saying
        # that the remote server should stop tracking a given path.
        remote_fs_observer = self.remote_fs_observer
        if remote_fs_observer is not None:
            self.remote_fs_observer = None
            writer = remote_fs_observer.writer
            if writer is not None:
                writer.write(
                    {"command": "stop_tracking", "on_change_id": self.on_change_id}
                )

            # Also remove from the parent.
            remote_fs_observer._change_id_to_fs_watch.pop(self.on_change_id, None)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IFSWatch = check_implements(self)


class RemoteFSObserver(object):
    def __init__(self, backend: str, extensions: Optional[Tuple[str, ...]]):
        import subprocess
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamReader
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamWriter

        self.process: Optional[subprocess.Popen] = None
        self.port: Optional[int] = None
        self.reader: Optional[JsonRpcStreamReader] = None
        self.writer: Optional[JsonRpcStreamWriter] = None
        self.reader_thread: Optional[threading.Thread] = None

        self._counter: "partial[int]" = partial(next, itertools.count())
        self._change_id_to_fs_watch: Dict[str, _RemoteFSWatch] = {}

        self._socket: Optional[socket_module.socket] = None
        self._backend = backend
        self._extensions = extensions
        self._initialized_event = threading.Event()

    def _next_id(self) -> str:
        return f"{os.getpid()} - {self._counter()} - {random.random()}"

    def start_server(
        self, log_file: Optional[str] = None, verbose: Optional[int] = None
    ) -> int:
        """
        :return int:
            The port used by the server (which may later be used to connect
            from another observer through connect_to_server).
        """
        assert self.port is None, "RemoteFSObserver already initialized."
        import subprocess
        from robocorp_ls_core import remote_fs_observer__main__

        args = [sys.executable, "-u", remote_fs_observer__main__.__file__]
        if log_file:
            args.append(f"--log-file={log_file}")

        if verbose:
            args.append("-" + ("v" * verbose))

        log.info("Initializing Remote FS Observer with the following args: %s", args)
        process = self.process = subprocess.Popen(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE
        )
        stdout = process.stdout
        assert stdout

        port_out = []
        read_port_thread_finished = threading.Event()

        def read_port():
            try:
                retcode = process.poll()

                while retcode is None:
                    contents = stdout.readline().strip()
                    if not contents:
                        log.info(
                            "Read empty string from Remote FS Observer process launched with: %s",
                            args,
                        )
                        time.sleep(0.1)
                        retcode = process.poll()
                    else:
                        if not contents.startswith(b"port:"):
                            log.critical(
                                f'Expected the read contents from Remote FS Observer to start with "port:". Found: {contents!r}'
                            )
                            continue

                        contents = contents[5:].strip()
                        port = int(contents)
                        port_out.append(port)
                        return

                # If it got here the process exited (and we didn't get the port).
                log.critical(
                    f"The Remote FS Observer process launched with: {args} exited without providing a port. retcode: {retcode}"
                )
            except:
                log.exception("Error reading port from Remote FS Observer.")
            finally:
                read_port_thread_finished.set()

        def read_fs_observer_stderr():
            while True:
                line = process.stderr.readline()
                if not line:
                    break
                line = line.decode("utf-8", "replace")
                sys.stderr.write(line)
                log.debug("Remote FS observer stderr: %s", line)

        t = threading.Thread(target=read_fs_observer_stderr)
        t.daemon = True
        t.start()

        t = threading.Thread(target=read_port)
        t.daemon = True
        t.start()

        if not read_port_thread_finished.wait(15):
            raise AssertionError(
                "Unable to read port from Remote FS Observer after 15 seconds."
            )

        if not port_out:
            raise AssertionError(
                "Remote FS Observer finished without providing a port."
            )

        port = self.port = port_out[0]
        self._initialize_reader_and_writer()
        assert self.writer, "Writer not properly initialized!"
        self.writer.write(
            {
                "command": "initialize",
                "backend": self._backend,
                "extensions": self._extensions,
                "parent_pid": os.getpid(),
            }
        )
        if not self._initialized_event.wait(5):
            raise RuntimeError("Unable to initialize server.")
        return port

    def connect_to_server(self, port: int):
        assert self.port is None, "RemoteFSObserver already initialized."

        self.port = port
        self._initialize_reader_and_writer()
        assert self.writer, "Writer not properly initialized!"
        self.writer.write({"command": "initialize_connect"})  # just used for the ack.
        if not self._initialized_event.wait(5):
            raise RuntimeError("Unable to initialize server.")

    def _initialize_reader_and_writer(self):
        assert self.writer is None
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamWriter
        from robocorp_ls_core.jsonrpc.streams import JsonRpcStreamReader

        s = socket_module.socket(socket_module.AF_INET, socket_module.SOCK_STREAM)
        try:
            # Configure the socket accordingly so that the connection is
            # properly kept alive.
            IPPROTO_TCP, SO_KEEPALIVE, TCP_KEEPIDLE, TCP_KEEPINTVL, TCP_KEEPCNT = (
                socket_module.IPPROTO_TCP,
                socket_module.SO_KEEPALIVE,
                socket_module.TCP_KEEPIDLE,  # @UndefinedVariable
                socket_module.TCP_KEEPINTVL,  # @UndefinedVariable
                socket_module.TCP_KEEPCNT,  # @UndefinedVariable
            )
            s.setsockopt(socket_module.SOL_SOCKET, SO_KEEPALIVE, 1)
            s.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
            s.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 3)
            s.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 5)
        except AttributeError:
            pass  # May not be available everywhere.

        s.connect(("127.0.0.1", self.port))
        self._socket = s
        write_to = s.makefile("wb")
        read_from = s.makefile("rb")

        w = JsonRpcStreamWriter(write_to, sort_keys=True)
        r = JsonRpcStreamReader(read_from)
        self.writer = w
        self.reader = r

        self.reader_thread = threading.Thread(
            target=self.reader.listen, args=(self._on_read,)
        )
        self.reader_thread.daemon = True
        self.reader_thread.start()

    def dispose(self):
        s = self._socket
        if s is not None:
            try:
                s.shutdown()
                s.close()
            except Exception:
                pass
            self._socket = None
        process = self.process
        if process:
            self.process.kill()
            self.process = None

    def notify_on_any_change(
        self,
        paths: List[PathInfo],
        on_change: IFSCallback,
        call_args=(),
        extensions: Optional[Sequence[str]] = None,
    ) -> IFSWatch:
        assert (
            self._initialized_event.is_set()
        ), "Initialization not completed. Unable to notify on changes."

        writer = self.writer
        if writer is None:
            raise RuntimeError(
                "Server communication is not initialized. start_server() or connect_to_server() must be called before notifying of changes."
            )

        on_change_id: str = self._next_id()
        path_args = [{"path": p.path, "recursive": p.recursive} for p in paths]

        remote_fs_watch = _RemoteFSWatch(self, on_change_id, on_change, call_args)
        self._change_id_to_fs_watch[on_change_id] = remote_fs_watch

        if extensions is not None:
            extensions = tuple(extensions)

        writer.write(
            {
                "command": "notify_on_any_change",
                "paths": path_args,
                "on_change_id": on_change_id,
                "extensions": extensions,
            }
        )
        # Wait for the command to be acknowledged...
        if not remote_fs_watch.acknowledged.wait(5):
            log.info(f"Command to notify on change not acknowledged. Paths: {paths}!")
        return remote_fs_watch

    def _on_read(self, msg):
        command = msg.get("command")

        if command == "ack_initialize":
            self._initialized_event.set()

        elif command == "ack_notify_on_any_change":
            on_change_id = msg["on_change_id"]
            remote_fs_watch: _RemoteFSWatch = self._change_id_to_fs_watch.get(
                on_change_id
            )
            if remote_fs_watch is not None:
                remote_fs_watch.acknowledged.set()
            else:
                log.info(
                    f"ack_notify_on_any_change did not find on_change_id: {on_change_id}"
                )

        elif command == "on_change":
            on_change_id = msg["on_change_id"]
            remote_fs_watch = self._change_id_to_fs_watch.get(on_change_id)
            # It may be None and that's ok (it may've been disposed in the meanwhile).
            if remote_fs_watch is not None:
                src_path = msg["src_path"]
                remote_fs_watch.on_change(src_path, *remote_fs_watch.call_args)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements
        from robocorp_ls_core.watchdog_wrapper import IFSObserver

        _: IFSObserver = check_implements(self)



from typing import List
from robocorp_ls_core.lsp import CompletionItemTypedDict
from robotframework_ls.impl.protocols import ICompletionContext

_SNIPPETS_RF4 = {
    "FOR IN": {
        "prefix": "FOR IN",
        "body": [
            "FOR    ${${1:element}}    IN    @{${2:LIST}}",
            "    Log    ${${1:element}}",
            "    $0",
            "END",
        ],
        "description": "Snippet of a FOR IN loop.\n\nA for loop that iterates over a list of values and assigns one value to a variable per iteration.",
    },
    "FOR IN ENUMERATE": {
        "prefix": "FOR IN ENUMERATE",
        "body": [
            "FOR    ${${1:index}}    ${${2:element}}    IN ENUMERATE    @{${3:LIST}}",
            "    Log    ${${1:index}}: ${${2:element}}",
            "    $0",
            "END",
        ],
        "description": "Snippet of a FOR IN ENUMERATE loop.\n\nA for loop that iterates over a list of values and assigns the iteration index to the first and the value to the second variable per iteration.",
    },
    "FOR IN RANGE": {
        "prefix": "FOR IN RANGE",
        "body": [
            "FOR    ${${1:counter}}    IN RANGE    ${2:START}    ${3:END}    ${4:opt.STEPS}",
            "    Log    ${${1:counter}}",
            "    $0",
            "END",
        ],
        "description": "Snippet of a FOR IN RANGE loop.\n\nA for loop that iterates over a range of values with an optional configurable step width.",
    },
    "FOR IN ZIP": {
        "prefix": "FOR IN ZIP",
        "body": [
            "FOR    ${${1:l1-element}}    ${${2:l2-element}}    IN ZIP    ${${3:LIST-1}}    ${${4:LIST-2}}",
            "    Log    ${${1:l1-element}} - ${${2:l2-element}}",
            "    $0",
            "END",
        ],
        "description": "Snippet of a FOR IN ZIP loop\n\nA for loop that iterates over two lists and assigns the values from the first list to the first variable and values from the second list to the second variable per iteration.",
    },
    "IF STATEMENT": {
        "prefix": "IF STATEMENT",
        "body": ["IF    ${${1:var1}} == ${${1:var2}}", "    $0", "END"],
        "description": "Snippet of an IF..END statement.",
    },
    "IF ELSE STATEMENT": {
        "prefix": "IF ELSE STATEMENT",
        "body": [
            "IF    ${${1:var1}} == ${${1:var2}}",
            "    ${3:Call Keyword}",
            "ELSE",
            "    $0",
            "END",
        ],
        "description": "Snippet of an IF..ELSE..END statement",
    },
    "Run Keyword If": {
        "prefix": "Run Keyword If",
        "body": [
            "Run Keyword If    ${1:condition}",
            "...    ${3:Keyword}    ${4:@args}",
            "...  ELSE IF    ${2:condition}",
            "...    ${5:Keyword}    ${6:@args}",
            "...  ELSE",
            "...    ${7:Keyword}    ${8:@args}",
        ],
        "description": "Runs the given keyword with the given arguments, if condition is true.",
    },
    "Run Keywords": {
        "prefix": "Run Keywords",
        "body": [
            "Run Keywords",
            "...    ${1:Keyword}    ${2:@args}",
            "...  AND",
            "...    ${3:Keyword}    ${4:@args}",
        ],
        "description": "Executes all the given keywords in a sequence.",
    },
}

_SNIPPETS_RF5 = {
    "TRY EXCEPT STATEMENT": {
        "prefix": "TRY EXCEPT",
        "body": ["TRY", "    $0", "EXCEPT  message", "    ", "END"],
        "description": "Snippet of a TRY..EXCEPT statement",
    },
    "TRY EXCEPT FINALLY STATEMENT": {
        "prefix": "TRY EXCEPT FINALLY",
        "body": ["TRY", "    $0", "EXCEPT  message", "    ", "FINALLY", "    ", "END"],
        "description": "Snippet of a TRY..EXCEPT..FINALLY statement",
    },
    "TRY FINALLY STATEMENT": {
        "prefix": "TRY FINALLY",
        "body": ["TRY", "    $0", "FINALLY", "    ", "END"],
        "description": "Snippet of a TRY..EXCEPT..FINALLY statement",
    },
    "WHILE STATEMENT": {
        "prefix": "WHILE",
        "body": [r"WHILE  ${1:expression}", "    $0", "END"],
        "description": "Snippet of a WHILE statement",
    },
}

_SNIPPETS_SORTED = None


def _get_global_snippets():
    from robotframework_ls.impl.robot_version import get_robot_major_version

    global _SNIPPETS_SORTED
    if _SNIPPETS_SORTED is None:
        use = {}
        use.update(_SNIPPETS_RF4)

        if get_robot_major_version() >= 5:
            use.update(_SNIPPETS_RF5)

        _SNIPPETS_SORTED = sorted(use.items())

    return _SNIPPETS_SORTED


def _create_completion_item_from_snippet(label, snippet, selection, line_to_col):
    """
    :param selection: DocumentSelection
    """
    from robocorp_ls_core.lsp import (
        CompletionItem,
        InsertTextFormat,
        Position,
        Range,
        TextEdit,
    )
    from robocorp_ls_core.lsp import CompletionItemKind

    current_col = selection.col

    text = "\n".join(snippet["body"])

    text_edit = TextEdit(
        Range(
            start=Position(selection.line, current_col - len(line_to_col)),
            end=Position(selection.line, current_col),
        ),
        text,
    )

    return CompletionItem(
        label,
        kind=CompletionItemKind.Snippet,
        text_edit=text_edit,
        insertText=text_edit.newText,
        documentation=snippet["description"] + "\n".join(["", ""] + snippet["body"]),
        insertTextFormat=InsertTextFormat.Snippet,
    ).to_dict()


def complete(completion_context: ICompletionContext) -> List[CompletionItemTypedDict]:
    """
    Collects all the keywords that are available to the given completion_context.

    :param CompletionContext completion_context:
    """
    sel = completion_context.sel  #::type sel: DocumentSelection
    line_to_column = sel.line_to_column.lstrip().lower()
    if not line_to_column:
        return []

    ret = []
    for label, data in _get_global_snippets():
        if line_to_column in data["prefix"].lower():
            ret.append(
                _create_completion_item_from_snippet(label, data, sel, line_to_column)
            )

    return ret



def generator2():
    for i in range(4):
        yield i


def generator():
    a = 42  # break here
    for x in generator2():
        yield x


sum = 0
for i in generator():
    sum += i

print('TEST SUCEEDED!')



import sys
import os


import unittest

class Test(unittest.TestCase):

    def test_it(self):
        #make it as if we were executing from the directory above this one (so that we can use jycompletionserver
        #without the need for it being in the pythonpath)
        #(twice the dirname to get the previous level from this file.)
        import test_pydevdio #@UnresolvedImport - importing itself
        ADD_TO_PYTHONPATH = os.path.join(os.path.dirname(os.path.dirname(test_pydevdio.__file__)))
        sys.path.insert(0, ADD_TO_PYTHONPATH)

        try:
            from _pydevd_bundle import pydevd_io
            original = sys.stdout

            try:
                sys.stdout = pydevd_io.IOBuf()
                print('foo')
                print('bar')

                self.assertEqual('foo\nbar\n', sys.stdout.getvalue()) #@UndefinedVariable

                print('ww')
                print('xx')
                self.assertEqual('ww\nxx\n', sys.stdout.getvalue()) #@UndefinedVariable
            finally:
                sys.stdout = original
        finally:
            #remove it to leave it ok for other tests
            sys.path.remove(ADD_TO_PYTHONPATH)




import os
import ast
from typing import List, Optional, Iterable
import difflib

from packaging import version
from robot.version import VERSION as RF_VERSION
from robot.api.parsing import ModelVisitor, Token, If, IfHeader, End
from robot.parsing.model import Statement
from robot.utils.robotio import file_writer
from click import style


ROBOT_VERSION = version.parse(RF_VERSION)


class StatementLinesCollector(ModelVisitor):
    """
    Used to get writeable presentation of Robot Framework model.
    """

    def __init__(self, model):
        self.text = ""
        self.visit(model)

    def visit_Statement(self, node):  # noqa
        for token in node.tokens:
            self.text += token.value

    def __eq__(self, other):
        return other.text == self.text


class GlobalFormattingConfig:
    def __init__(
        self,
        space_count: int,
        line_sep: str,
        start_line: Optional[int],
        end_line: Optional[int],
        separator: str,
        line_length: int,
    ):
        self.start_line = start_line
        self.end_line = end_line
        self.space_count = space_count
        self.line_length = line_length

        if separator == "space":
            self.separator = " " * space_count
        elif separator == "tab":
            self.space_count = space_count
            self.separator = "\t"

        if line_sep == "windows":
            self.line_sep = "\r\n"
        elif line_sep == "unix":
            self.line_sep = "\n"
        elif line_sep == "auto":
            self.line_sep = "auto"
        else:
            self.line_sep = os.linesep


def decorate_diff_with_color(contents: List[str]) -> str:
    """Inject the ANSI color codes to the diff."""
    for i, line in enumerate(contents):
        if line.startswith("+++") or line.startswith("---"):
            line = style(line, bold=True, reset=True)
        elif line.startswith("@@"):
            line = style(line, fg="cyan", reset=True)
        elif line.startswith("+"):
            line = style(line, fg="green", reset=True)
        elif line.startswith("-"):
            line = style(line, fg="red", reset=True)
        contents[i] = line
    return "".join(contents)


def normalize_name(name):
    return name.lower().replace("_", "").replace(" ", "")


def after_last_dot(name):
    return name.split(".")[-1]


def node_within_lines(node_start, node_end, start_line, end_line):
    if start_line:
        if node_start < start_line:
            return False
        if end_line:
            if node_end > end_line:
                return False
        else:
            if start_line != node_start:
                return False
    return True


def node_outside_selection(node, formatting_config):
    """
    Contrary to ``node_within_lines`` it just checks if node is fully outside selected lines.
    Partial selection is useful for transformers like aligning code.
    """
    if (
        formatting_config.start_line
        and formatting_config.start_line > node.end_lineno
        or formatting_config.end_line
        and formatting_config.end_line < node.lineno
    ):
        return True
    return False


def split_args_from_name_or_path(name):
    """Split arguments embedded to name or path like ``Example:arg1:arg2``.

    The separator can be either colon ``:`` or semicolon ``;``. If both are used,
    the first one is considered to be the separator.
    """
    if os.path.exists(name):
        return name, []
    index = _get_arg_separator_index_from_name_or_path(name)
    if index == -1:
        return name, []
    args = _escaped_split(name[index + 1 :], name[index])
    name = name[:index]
    return name, args


def _escaped_split(string, delim):
    ret = []
    current = []
    itr = iter(string)
    for ch in itr:
        if ch == "\\":
            try:
                current.append("\\")
                current.append(next(itr))
            except StopIteration:
                pass
        elif ch == delim:
            ret.append("".join(current))
            current = []
        else:
            current.append(ch)
    if current:
        ret.append("".join(current))
    return ret


def _get_arg_separator_index_from_name_or_path(name):
    colon_index = name.find(":")
    # Handle absolute Windows paths
    if colon_index == 1 and name[2:3] in ("/", "\\"):
        colon_index = name.find(":", colon_index + 1)
    semicolon_index = name.find(";")
    if colon_index == -1:
        return semicolon_index
    if semicolon_index == -1:
        return colon_index
    return min(colon_index, semicolon_index)


def round_to_four(number):
    div = number % 4
    if div:
        return number + 4 - div
    return number


def any_non_sep(tokens):
    return any(token.type not in (Token.EOL, Token.SEPARATOR, Token.EOS) for token in tokens)


def tokens_by_lines(node):
    for line in node.lines:
        if not any_non_sep(line):
            continue
        if line:
            if line[0].type == Token.VARIABLE:
                if line[0].value:
                    line[0].value = line[0].value.lstrip()
                else:
                    # if variable is prefixed with spaces
                    line = line[1:]
            elif line[0].type == Token.ARGUMENT:
                line[0].value = line[0].value.strip() if line[0].value else line[0].value
        yield [token for token in line if token.type not in (Token.SEPARATOR, Token.EOS)]


def left_align(node):
    """remove leading separator token"""
    tokens = list(node.tokens)
    if tokens:
        tokens[0].value = tokens[0].value.lstrip(" \t")
    return Statement.from_tokens(tokens)


def remove_rst_formatting(text):
    return text.replace("::", ":").replace("``", "'")


class RecommendationFinder:
    def find_similar(self, name, candidates):
        norm_name = name.lower()
        norm_cand = self.get_normalized_candidates(candidates)
        matches = self.find(norm_name, norm_cand.keys())
        if not matches:
            return ""
        matches = self.get_original_candidates(matches, norm_cand)
        if len(matches) == 1 and matches[0] == name:
            return ""
        suggestion = " Did you mean:\n"
        suggestion += "\n".join(f"    {match}" for match in matches)
        return suggestion

    def find(self, name, candidates, max_matches=2):
        """Return a list of close matches to `name` from `candidates`."""
        if not name or not candidates:
            return []
        cutoff = self._calculate_cutoff(name)
        return difflib.get_close_matches(name, candidates, n=max_matches, cutoff=cutoff)

    @staticmethod
    def _calculate_cutoff(string, min_cutoff=0.5, max_cutoff=0.85, step=0.03):
        """The longer the string the bigger required cutoff."""
        cutoff = min_cutoff + len(string) * step
        return min(cutoff, max_cutoff)

    @staticmethod
    def get_original_candidates(candidates, norm_candidates):
        """Map found normalized candidates to unique original candidates."""
        return sorted(list(set(c for cand in candidates for c in norm_candidates[cand])))

    @staticmethod
    def get_normalized_candidates(candidates):
        norm_cand = {cand.lower(): [cand] for cand in candidates}
        # most popular typos
        norm_cand["align"] = ["AlignSettingsSection", "AlignVariablesSection"]
        norm_cand["normalize"] = [
            "NormalizeAssignments",
            "NormalizeNewLines",
            "NormalizeSectionHeaderName",
            "NormalizeSeparators",
            "NormalizeSettingName",
        ]
        norm_cand["order"] = ["OrderSettings", "OrderSettingsSection"]
        norm_cand["alignsettings"] = ["AlignSettingsSection"]
        norm_cand["alignvariables"] = ["AlignVariablesSection"]
        norm_cand["assignmentnormalizer"] = ["NormalizeAssignments"]
        return norm_cand


class ModelWriter(ModelVisitor):
    def __init__(self, output, newline):
        self.writer = file_writer(output, newline=newline)
        self.close_writer = True

    def write(self, model):
        try:
            self.visit(model)
        finally:
            if self.close_writer:
                self.writer.close()

    def visit_Statement(self, statement):  # noqa
        for token in statement.tokens:
            self.writer.write(token.value)


class TestTemplateFinder(ast.NodeVisitor):
    def __init__(self):
        self.templated = False

    def visit_TestTemplate(self, node):  # noqa
        if node.value:
            self.templated = True


def is_suite_templated(node):
    template_finder = TestTemplateFinder()
    template_finder.visit(node)
    return template_finder.templated


def is_blank_multiline(statements):
    return (
        statements[0].type == Token.CONTINUATION
        and len(statements) == 3
        and statements[1].type == "ARGUMENT"
        and not statements[1].value
    )


def replace_indent_in_lines(node, indent_len):
    tokens = []
    for line in node.lines:
        if line and line[0].type == Token.SEPARATOR:
            indent_token = Token(Token.SEPARATOR, len(line[0].value) * " " + indent_len)
            line[0] = indent_token
        tokens.extend(line)
    return tokens


def create_statement_from_tokens(statement, tokens: Iterable, indent: Token):
    return statement([indent, Token(statement.type), *tokens])


def wrap_in_if_and_replace_statement(node, statement, default_separator):
    if len(node.data_tokens) < 2:
        return node
    condition = node.data_tokens[1]
    indent = Token(Token.SEPARATOR, node.tokens[0].value + default_separator)
    indented_tokens = replace_indent_in_lines(node, default_separator)
    body = create_statement_from_tokens(statement=statement, tokens=indented_tokens[4:], indent=indent)
    header = IfHeader(
        [
            node.tokens[0],
            Token(Token.IF),
            Token(Token.SEPARATOR, default_separator),
            condition,
            Token(Token.EOL),
        ]
    )
    end = End.from_params(indent=node.tokens[0].value)
    if_block = If(header=header, body=[body], orelse=None, end=end)
    return if_block


def get_comments(tokens):
    prev_sep = ""
    comments = []
    for token in tokens:
        if token.type == Token.COMMENT:
            if token.value.startswith("#"):
                comments.append(token)
            elif comments:
                comments[-1].value += prev_sep + token.value
            else:
                token.value = f"# {token.value}"
                comments.append(token)
        elif token.type == Token.SEPARATOR:
            prev_sep = token.value
    return comments


def flatten_multiline(tokens, separator, remove_comments: bool = False):
    flattened = []
    skip_start = False
    for tok in tokens[:-1]:
        if tok.type == Token.EOL:
            skip_start = True
        elif skip_start:
            if tok.type == Token.CONTINUATION:
                skip_start = False
        else:
            if tok.type == Token.ARGUMENT and tok.value == "":
                flattened.append(Token(Token.SEPARATOR, separator))
                tok.value = "${EMPTY}"
            if remove_comments and tok.type == Token.COMMENT:
                if flattened and flattened[-1].type == Token.SEPARATOR:
                    flattened.pop()
            else:
                flattened.append(tok)
    flattened.append(tokens[-1])
    return flattened



from __future__ import annotations

import queue
import threading
from typing import Optional
import typing

from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import BaseSchema
from robocorp_ls_core.options import DEFAULT_TIMEOUT
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_debug_adapter.base_launch_process_target import (
    BaseLaunchProcessTargetComm,
    IProtocolMessageCallable,
)


if typing.TYPE_CHECKING:
    from robotframework_debug_adapter.debug_adapter_comm import DebugAdapterComm

log = get_logger(__name__)


class LaunchProcessDebugAdapterPydevdComm(BaseLaunchProcessTargetComm):
    def __init__(self, debug_adapter_comm: DebugAdapterComm, server_socket):
        BaseLaunchProcessTargetComm.__init__(self, debug_adapter_comm)
        self._server_socket = server_socket
        assert server_socket is not None
        self._write_to_pydevd_queue: "queue.Queue[BaseSchema]" = queue.Queue()

    def run(self):
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            writer_thread_no_auto_seq,
        )
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            reader_thread,
        )

        try:

            # while True:
            # Only handle a single connection...
            socket, _addr = self._server_socket.accept()

            read_from = socket.makefile("rb")
            write_to = socket.makefile("wb")

            debug_adapter_comm = self._weak_debug_adapter_comm()
            writer = self._writer_thread = threading.Thread(
                target=writer_thread_no_auto_seq,
                args=(write_to, self._write_to_pydevd_queue, "write to pydevd process"),
                name="Write to pydevd (LaunchProcessDebugAdapterPydevdComm)",
            )
            writer.daemon = True

            reader = self._reader_thread = threading.Thread(
                target=reader_thread,
                args=(
                    read_from,
                    self._from_pydevd,
                    debug_adapter_comm.write_to_client_queue,  # Used for errors
                    b"read from pydevd process",
                ),
                name="Read from pydevd (LaunchProcessDebugAdapterPydevdComm)",
            )
            reader.daemon = True

            reader.start()
            writer.start()

            self._connected_event.set()
        except:
            log.exception()

    def _from_pydevd(self, protocol_message: BaseSchema) -> None:
        self._handle_received_protocol_message_from_backend(protocol_message, "pydevd")

    def write_to_pydevd_message(
        self, protocol_message, on_response: Optional[IProtocolMessageCallable] = None
    ):
        """
        :param BaseSchema protocol_message:
            Some instance of one of the messages in the debug_adapter.schema.
        """
        seq = protocol_message.seq = self._next_seq()
        if on_response is not None:
            self._msg_id_to_on_response[seq] = on_response
        self._write_to_pydevd_queue.put(protocol_message)

    def wait_for_connection(self):
        """
        :return bool:
            Returns True if the connection was successful and False otherwise.
        """
        assert self.is_alive() is not None, "start() must be called first."
        log.debug("Wating for connection for %s seconds." % (DEFAULT_TIMEOUT,))
        ret = self._connected_event.wait(DEFAULT_TIMEOUT)
        log.debug("Connected: %s" % (ret,))
        return ret



from robotframework_ls.impl.protocols import ICompletionContext
from typing import Optional


class _Requisites(object):
    def __init__(self, section, matcher, replace_from_col, replace_to_col, selection):
        self.section = section
        self.matcher = matcher
        self.replace_from_col = replace_from_col
        self.replace_to_col = replace_to_col
        self.selection = selection


def get_requisites(completion_context: ICompletionContext) -> Optional[_Requisites]:
    section_name = completion_context.get_current_section_name()
    if not section_name:
        return None

    from robotframework_ls.impl.string_matcher import RobotStringMatcher

    section = completion_context.get_section(section_name)
    if section is None:
        return None

    selection = completion_context.sel  #: :type selection: DocumentSelection
    line_to_col = selection.line_to_column
    if line_to_col.endswith("  "):
        return None
    replace_to_col = selection.col
    if section.names_in_brackets:
        for i, c in enumerate(line_to_col):
            if c.isspace():
                continue
            elif c == "[":
                line_to_col = line_to_col[i + 1 :]
                replace_from_col = i
                break
            else:
                return None
        else:
            return None

        matcher = RobotStringMatcher(line_to_col)

    else:
        # i.e.: Needs to be the first char
        matcher = RobotStringMatcher(line_to_col)
        replace_from_col = 0

    return _Requisites(section, matcher, replace_from_col, replace_to_col, selection)


def complete(completion_context: ICompletionContext):
    import itertools
    from robocorp_ls_core.lsp import (
        TextEdit,
        Range,
        Position,
        CompletionItem,
        CompletionItemKind,
    )

    requisites = get_requisites(completion_context)
    if requisites is None:
        return []

    section = requisites.section
    matcher = requisites.matcher
    replace_from_col = requisites.replace_from_col
    replace_to_col = requisites.replace_to_col
    selection = requisites.selection

    ret = []
    for word in sorted(itertools.chain(section.names, section.aliases)):
        if matcher.accepts(word):
            if section.names_in_brackets:
                label = "[%s]" % (word,)
                line = selection.current_line
                replacement = "[%s]" % (word,)
                if line[selection.col :].startswith("]"):
                    replace_to_col += 1

            else:
                label = word
                replacement = word

            text_edit = TextEdit(
                Range(
                    start=Position(selection.line, replace_from_col),
                    end=Position(selection.line, replace_to_col),
                ),
                replacement,
            )
            # text_edit = None
            ret.append(
                CompletionItem(
                    label, kind=CompletionItemKind.Keyword, text_edit=text_edit
                ).to_dict()
            )

    return ret



def Method1():
    print('m1')
    print('m1')
    
def Method1a():
    print('m1a')
    print('m1a')

def Method2():
    print('m2 before')
    Method1()
    Method1a()
    print('m2 after')

   
if __name__ == '__main__': 
    Method2()
    print('TEST SUCEEDED!')



# Original work Copyright 2018 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import traceback


class JsonRpcException(Exception):
    def __init__(self, message=None, code=None, data=None):
        super(JsonRpcException, self).__init__(message)
        self.message = message or getattr(self.__class__, "MESSAGE")
        self.code = code or getattr(self.__class__, "CODE")
        self.data = data

    def to_dict(self):
        exception_dict = {"code": self.code, "message": self.message}
        if self.data is not None:
            exception_dict["data"] = self.data
        return exception_dict

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.code == other.code
            and self.message == other.message
        )

    def __hash__(self):
        return hash((self.code, self.message))

    @staticmethod
    def from_dict(error):
        for exc_class in _EXCEPTIONS:
            if exc_class.supports_code(error["code"]):
                return exc_class(**error)
        return JsonRpcException(**error)

    @classmethod
    def supports_code(cls, code):
        # Defaults to UnknownErrorCode
        return getattr(cls, "CODE", -32001) == code


class JsonRpcParseError(JsonRpcException):
    CODE = -32700
    MESSAGE = "Parse Error"


class JsonRpcInvalidRequest(JsonRpcException):
    CODE = -32600
    MESSAGE = "Invalid Request"


class JsonRpcMethodNotFound(JsonRpcException):
    CODE = -32601
    MESSAGE = "Method Not Found"

    @classmethod
    def of(cls, method):
        return cls(message=cls.MESSAGE + ": " + method)


class JsonRpcInvalidParams(JsonRpcException):
    CODE = -32602
    MESSAGE = "Invalid Params"


class JsonRpcInternalError(JsonRpcException):
    CODE = -32602
    MESSAGE = "Internal Error"

    @classmethod
    def of(cls, exc_info):
        exc_type, exc_value, exc_tb = exc_info
        return cls(
            message="".join(
                traceback.format_exception_only(exc_type, exc_value)
            ).strip(),
            data={"traceback": traceback.format_tb(exc_tb)},
        )


class JsonRpcRequestCancelled(JsonRpcException):
    CODE = -32800
    MESSAGE = "Request Cancelled"


class JsonRpcServerError(JsonRpcException):
    def __init__(self, message, code, data=None):
        assert _is_server_error_code(code)
        super(JsonRpcServerError, self).__init__(message=message, code=code, data=data)

    @classmethod
    def supports_code(cls, code):
        return _is_server_error_code(code)


def _is_server_error_code(code):
    return -32099 <= code <= -32000


_EXCEPTIONS = (
    JsonRpcParseError,
    JsonRpcInvalidRequest,
    JsonRpcMethodNotFound,
    JsonRpcInvalidParams,
    JsonRpcInternalError,
    JsonRpcRequestCancelled,
    JsonRpcServerError,
)



from functools import partial
from typing import List, Optional

from robocorp_ls_core.basic import log_and_silence_errors
from robocorp_ls_core.jsonrpc.endpoint import require_monitor
from robocorp_ls_core.protocols import (
    IRobotFrameworkApiClient,
    IMonitor,
    IIdMessageMatcher,
)
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.constants import DEFAULT_COMPLETIONS_TIMEOUT
from robocorp_ls_core.lsp import CompletionItemTypedDict


log = get_logger(__name__)


class _RobotFrameworkLsCompletionImpl(object):
    def __init__(self, server_manager, language_server):
        import weakref

        self._weak_robot_framework_ls = weakref.ref(language_server)
        self._server_manager = server_manager

        self._last_stored = {}

    def text_document_completion(self, doc_uri: str, line: int, col: int):
        rf_api_client = self._server_manager.get_regular_rf_api_client(doc_uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_document_completion,
                rf_api_client,
                doc_uri,
                line,
                col,
            )
            func = require_monitor(func)
            return func

        log.info("Unable to get completions (no api available).")
        return []

    @log_and_silence_errors(log, return_on_error=[])
    def _threaded_document_completion(
        self,
        rf_api_client: IRobotFrameworkApiClient,
        doc_uri: str,
        line: int,
        col: int,
        monitor: IMonitor,
    ) -> list:
        from robotframework_ls.impl.completion_context import CompletionContext
        from robotframework_ls.impl import section_completions
        from robocorp_ls_core.client_base import wait_for_message_matchers

        ls = self._weak_robot_framework_ls()
        if not ls:
            log.critical("RobotFrameworkLanguageServer garbage-collected.")
            return []

        ws = ls.workspace
        if not ws:
            log.critical("Workspace must be set before returning completions.")
            return []

        document = ws.get_document(doc_uri, accept_from_file=True)
        if document is None:
            log.critical("Unable to find document (%s) for completions." % (doc_uri,))
            return []

        ctx = CompletionContext(document, line, col, config=ls.config, workspace=ws)
        completions = []

        # Asynchronous completion.
        message_matchers: List[Optional[IIdMessageMatcher]] = []
        message_matchers.append(rf_api_client.request_complete_all(doc_uri, line, col))

        # These run locally (no need to get from the server).
        completions.extend(section_completions.complete(ctx))

        accepted_message_matchers = wait_for_message_matchers(
            message_matchers,
            monitor,
            rf_api_client.request_cancel,
            DEFAULT_COMPLETIONS_TIMEOUT,
        )
        for message_matcher in accepted_message_matchers:
            msg = message_matcher.msg
            if msg is not None:
                result = msg.get("result")
                if result:
                    for completion_item in result:
                        data = completion_item.get("data")
                        if data and isinstance(data, dict):
                            # We need to add the doc uri to the data.
                            data["uri"] = doc_uri
                        completions.append(completion_item)

        return completions

    def resolve_completion_item(self, completion_item: CompletionItemTypedDict):
        try:
            data = completion_item.get("data")
            if data and isinstance(data, dict):
                doc_uri = data.get("uri")
                ls = self._weak_robot_framework_ls()
                if not ls:
                    log.critical(
                        "RobotFrameworkLanguageServer garbage-collected in resolve completion item."
                    )
                    return completion_item

                return ls.async_api_forward(
                    "request_resolve_completion_item",
                    "api",
                    doc_uri,
                    default_return=completion_item,
                    __add_doc_uri_in_args__=False,
                    completion_item=completion_item,
                )
        except:
            log.exception("Error resolving completion item.")

        return completion_item



import sys
import os
from typing import (
    Optional,
    Any,
    List,
    Tuple,
    Set,
    Callable,
    Dict,
    Iterator,
    Sequence,
)

from robocorp_ls_core.cache import instance_cache
from robocorp_ls_core.constants import NULL
from robocorp_ls_core.lsp import CompletionItemTypedDict, MarkupContentTypedDict
from robocorp_ls_core.protocols import (
    IMonitor,
    Sentinel,
    IConfig,
    IDocumentSelection,
    IWorkspace,
)
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    IRobotDocument,
    ICompletionContext,
    TokenInfo,
    IRobotWorkspace,
    IKeywordDefinition,
    ILibraryImportNode,
    KeywordUsageInfo,
    CompletionType,
    IResourceImportNode,
    ICompletionContextDependencyGraph,
    IRobotToken,
    INode,
    IVariableImportNode,
    VarTokenInfo,
    IVariablesFromArgumentsFileLoader,
    IVariableFound,
    NodeInfo,
)
from robotframework_ls.impl.robot_workspace import RobotDocument
from robocorp_ls_core import uris
import itertools
from functools import partial
import typing


log = get_logger(__name__)


class _Memo(object):
    pass


class BaseContext(object):
    def __init__(self, workspace: IRobotWorkspace, config: IConfig, monitor: IMonitor):
        self._workspace = workspace
        self._config = config
        self._monitor = monitor

    @property
    def monitor(self) -> IMonitor:
        return self._monitor

    @property
    def workspace(self) -> IRobotWorkspace:
        return self._workspace

    @property
    def config(self) -> IConfig:
        return self._config

    def check_cancelled(self) -> None:
        self._monitor.check_cancelled()

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements
        from robotframework_ls.impl.protocols import IBaseCompletionContext

        _: IBaseCompletionContext = check_implements(self)


class CompletionContext(object):

    TYPE_TEST_CASE = RobotDocument.TYPE_TEST_CASE
    TYPE_INIT = RobotDocument.TYPE_INIT
    TYPE_RESOURCE = RobotDocument.TYPE_RESOURCE

    def __init__(
        self,
        doc,
        line=Sentinel.SENTINEL,
        col=Sentinel.SENTINEL,
        workspace: Optional[IWorkspace] = None,
        config: Optional[IConfig] = None,
        memo: Optional[_Memo] = None,
        monitor: Optional[IMonitor] = NULL,
        variables_from_arguments_files_loader: Sequence[
            IVariablesFromArgumentsFileLoader
        ] = (),
    ) -> None:
        if col is Sentinel.SENTINEL or line is Sentinel.SENTINEL:
            assert (
                col is Sentinel.SENTINEL
            ), "Either line and col are not set, or both are set. Found: (%s, %s)" % (
                line,
                col,
            )
            assert (
                line is Sentinel.SENTINEL
            ), "Either line and col are not set, or both are set. Found: (%s, %s)" % (
                line,
                col,
            )

            # If both are not set, use the doc len as the selection.
            line, col = doc.get_last_line_col()

        memo = _Memo() if memo is None else memo

        sel = doc.selection(line, col)

        self._doc = doc
        self._sel = sel
        self._workspace = typing.cast(Optional[IRobotWorkspace], workspace)
        self._config = config
        self._memo = memo
        self._original_ctx: Optional[CompletionContext] = None
        self._monitor = monitor or NULL
        self.type = CompletionType.regular

        self._id_to_compute_documentation: Dict[
            int, Callable[[], MarkupContentTypedDict]
        ] = {}
        self.variables_from_arguments_files_loader = (
            variables_from_arguments_files_loader
        )

    def assign_documentation_resolve(
        self,
        completion_item: CompletionItemTypedDict,
        compute_documentation: Callable[[], MarkupContentTypedDict],
    ) -> None:
        if self._original_ctx is not None:
            self._original_ctx.assign_documentation_resolve(
                completion_item, compute_documentation
            )
        else:
            next_id = len(self._id_to_compute_documentation)
            self._id_to_compute_documentation[next_id] = compute_documentation
            completion_item["data"] = {"id": next_id, "ctx": id(self)}

    def resolve_completion_item(
        self, data, completion_item: CompletionItemTypedDict, monaco=False
    ) -> None:
        if self._original_ctx is not None:
            self._original_ctx.resolve_completion_item(data, completion_item)
        else:
            compute_documentation = self._id_to_compute_documentation.get(
                data.get("id")
            )
            if compute_documentation is not None:
                marked: Optional[MarkupContentTypedDict] = compute_documentation()
                if marked:
                    if monaco:
                        if marked["kind"] == "markdown":
                            completion_item["documentation"] = {
                                "value": marked["value"]
                            }
                        else:
                            completion_item["documentation"] = marked["value"]
                    else:
                        completion_item["documentation"] = marked

    @property
    def monitor(self) -> IMonitor:
        return self._monitor

    def check_cancelled(self) -> None:
        self._monitor.check_cancelled()

    def create_copy_with_selection(self, line: int, col: int) -> ICompletionContext:
        doc = self._doc
        ctx = CompletionContext(
            doc,
            line=line,
            col=col,
            workspace=self._workspace,
            config=self._config,
            memo=self._memo,
            monitor=self._monitor,
            variables_from_arguments_files_loader=self.variables_from_arguments_files_loader,
        )
        ctx._original_ctx = self
        return ctx

    def create_copy(self, doc: IRobotDocument) -> ICompletionContext:
        ctx = CompletionContext(
            doc,
            line=0,
            col=0,
            workspace=self._workspace,
            config=self._config,
            memo=self._memo,
            monitor=self._monitor,
            variables_from_arguments_files_loader=self.variables_from_arguments_files_loader,
        )
        ctx._original_ctx = self
        return ctx

    @property
    def original_doc(self) -> IRobotDocument:
        if self._original_ctx is None:
            return self._doc
        return self._original_ctx.original_doc

    @property
    def original_sel(self) -> IDocumentSelection:
        if self._original_ctx is None:
            return self._sel
        return self._original_ctx.original_sel

    @property
    def doc(self) -> IRobotDocument:
        return self._doc

    @property
    def sel(self) -> IDocumentSelection:
        return self._sel

    @property
    def memo(self) -> Any:
        return self._memo

    @property
    def config(self) -> Optional[IConfig]:
        return self._config

    @property
    def workspace(self) -> IRobotWorkspace:
        assert self._workspace
        return self._workspace

    @instance_cache
    def get_type(self) -> Any:
        return self.doc.get_type()

    @instance_cache
    def get_ast(self) -> Any:
        return self.doc.get_ast()

    @instance_cache
    def get_ast_current_section(self) -> Any:
        """
        :rtype: robot.parsing.model.blocks.Section|NoneType
        """
        from robotframework_ls.impl import ast_utils

        ast = self.get_ast()
        section = ast_utils.find_section(ast, self.sel.line)
        return section

    def get_accepted_section_header_words(self) -> List[str]:
        """
        :rtype: list(str)
        """
        sections = self._get_accepted_sections()
        ret = []
        for section in sections:
            for marker in section.markers:
                ret.append(marker.title())
        ret.sort()
        return ret

    def get_current_section_name(self) -> Optional[str]:
        """
        :rtype: str|NoneType
        """
        section = self.get_ast_current_section()

        section_name = None
        header = getattr(section, "header", None)
        if header is not None:
            try:
                section_name = header.name
            except AttributeError:
                section_name = header.value  # older version of 3.2

        return section_name

    def _get_accepted_sections(self) -> list:
        """
        :rtype: list(robot_constants.Section)
        """
        from robotframework_ls.impl import robot_constants

        t = self.get_type()
        if t == self.TYPE_TEST_CASE:
            return robot_constants.TEST_CASE_FILE_SECTIONS

        elif t == self.TYPE_RESOURCE:
            return robot_constants.RESOURCE_FILE_SECTIONS

        elif t == self.TYPE_INIT:
            return robot_constants.INIT_FILE_SECTIONS

        else:
            log.critical("Unrecognized section: %s", t)
            return robot_constants.TEST_CASE_FILE_SECTIONS

    def get_section(self, section_name: str) -> Any:
        """
        :rtype: robot_constants.Section
        """
        section_name = section_name.lower()
        accepted_sections = self._get_accepted_sections()

        for section in accepted_sections:
            for marker in section.markers:
                if marker.lower() == section_name:
                    return section
        return None

    @instance_cache
    def get_current_token(self) -> Optional[TokenInfo]:
        from robotframework_ls.impl import ast_utils

        section = self.get_ast_current_section()
        if section is None:
            return None
        return ast_utils.find_token(section, self.sel.line, self.sel.col)

    def get_all_variables(self) -> Tuple[NodeInfo, ...]:
        from robotframework_ls.impl import ast_utils

        ast = self.get_ast()
        return tuple(ast_utils.iter_variables(ast))

    @instance_cache
    def get_doc_normalized_var_name_to_var_found(self) -> Dict[str, IVariableFound]:
        from robotframework_ls.impl import ast_utils
        from robotframework_ls.impl.variable_resolve import robot_search_variable
        from robot.api import Token
        from robotframework_ls.impl.variable_types import VariableFoundFromToken
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        ret: Dict[str, IVariableFound] = {}
        for variable_node_info in self.get_all_variables():
            variable_node = variable_node_info.node
            token = variable_node.get_token(Token.VARIABLE)
            if token is None:
                continue

            variable_match = robot_search_variable(token.value)
            # Filter out empty base
            if variable_match is None or not variable_match.base:
                continue

            base_token = ast_utils.convert_variable_match_base_to_token(
                token, variable_match
            )
            ret[normalize_robot_name(variable_match.base)] = VariableFoundFromToken(
                self,
                base_token,
                variable_node.value,
                variable_name=variable_match.base,
            )

        return ret

    @instance_cache
    def get_settings_normalized_var_name_to_var_found(
        self,
    ) -> Dict[str, IVariableFound]:
        from robotframework_ls.impl.text_utilities import normalize_robot_name
        from robotframework_ls.impl.variable_types import VariableFoundFromSettings

        ret: Dict[str, IVariableFound] = {}

        from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_VARIABLES

        config = self.config
        if config is not None:
            robot_variables = config.get_setting(OPTION_ROBOT_VARIABLES, dict, {})
            for key, val in robot_variables.items():
                ret[normalize_robot_name(key)] = VariableFoundFromSettings(key, val)

        return ret

    @instance_cache
    def get_builtins_normalized_var_name_to_var_found(
        self, resolved
    ) -> Dict[str, IVariableFound]:
        from robotframework_ls.impl.text_utilities import normalize_robot_name
        from robotframework_ls.impl.variable_types import VariableFoundFromBuiltins
        from robotframework_ls.impl.robot_constants import BUILTIN_VARIABLES_RESOLVED

        ret: Dict[str, IVariableFound] = {}

        from robotframework_ls.impl.robot_constants import get_builtin_variables

        for key, val in get_builtin_variables():
            ret[normalize_robot_name(key)] = VariableFoundFromBuiltins(key, val)

        if resolved:
            for key, val in BUILTIN_VARIABLES_RESOLVED.items():
                # Provide a resolved value for the ones we can resolve.
                ret[normalize_robot_name(key)] = VariableFoundFromBuiltins(key, val)

        return ret

    def get_arguments_files_normalized_var_name_to_var_found(
        self,
    ) -> Dict[str, IVariableFound]:
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        ret: Dict[str, IVariableFound] = {}

        if not self.variables_from_arguments_files_loader:
            return ret

        for c in self.variables_from_arguments_files_loader:
            for variable in c.get_variables():
                ret[normalize_robot_name(variable.variable_name)] = variable

        return ret

    @instance_cache
    def get_current_variable(self, section=None) -> Optional[VarTokenInfo]:
        """
        Provides the current variable token. Note that it won't include '{' nor '}'.
        """
        from robotframework_ls.impl import ast_utils

        if section is None:
            section = self.get_ast_current_section()

        if section is None:
            return None
        return ast_utils.find_variable(section, self.sel.line, self.sel.col)

    @instance_cache
    def get_imported_libraries(self) -> Tuple[ILibraryImportNode, ...]:
        from robotframework_ls.impl import ast_utils

        ast = self.get_ast()
        ret = []
        for library_import in ast_utils.iter_library_imports(ast):
            if library_import.node.name:
                ret.append(library_import.node)
        return tuple(ret)

    @instance_cache
    def get_resource_imports(self):
        from robotframework_ls.impl import ast_utils

        ast = self.get_ast()
        ret = []
        for resource in ast_utils.iter_resource_imports(ast):
            if resource.node.name:
                ret.append(resource.node)
        return tuple(ret)

    @instance_cache
    def get_variable_imports(self) -> Tuple[INode, ...]:
        from robotframework_ls.impl import ast_utils

        ast = self.get_ast()
        ret = []
        for resource in ast_utils.iter_variable_imports(ast):
            if resource.node.name:
                ret.append(resource.node)
        return tuple(ret)

    def token_value_resolving_variables(self, token: IRobotToken) -> str:
        from robotframework_ls.impl.variable_resolve import ResolveVariablesContext

        return ResolveVariablesContext(self).token_value_resolving_variables(token)

    def token_value_and_unresolved_resolving_variables(
        self, token: IRobotToken
    ) -> Tuple[str, Tuple[IRobotToken, ...]]:
        from robotframework_ls.impl.variable_resolve import ResolveVariablesContext

        return ResolveVariablesContext(
            self
        ).token_value_and_unresolved_resolving_variables(token)

    @instance_cache
    def get_resource_import_as_doc(
        self, resource_import: INode
    ) -> Optional[IRobotDocument]:
        from robot.api import Token
        from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_PYTHONPATH

        ws = self.workspace
        token = resource_import.get_token(Token.NAME)
        if token is not None:
            name_with_resolved_vars = self.token_value_resolving_variables(token)

            if not os.path.isabs(name_with_resolved_vars):
                # It's a relative resource, resolve its location based on the
                # current file.
                check_paths = [
                    os.path.normpath(
                        os.path.join(
                            os.path.dirname(self.doc.path), name_with_resolved_vars
                        )
                    )
                ]
                config = self.config
                if config is not None:
                    for additional_pythonpath_entry in config.get_setting(
                        OPTION_ROBOT_PYTHONPATH, list, []
                    ):
                        check_paths.append(
                            os.path.normpath(
                                os.path.join(
                                    additional_pythonpath_entry,
                                    name_with_resolved_vars,
                                )
                            )
                        )

                for path in sys.path:
                    check_paths.append(
                        os.path.normpath(
                            os.path.join(
                                path,
                                name_with_resolved_vars,
                            )
                        )
                    )

                # Make the entries in the list unique (keeping order).
                seen: Set[str] = set()
                add_it_and_return_none = seen.add
                check_paths = [
                    x
                    for x in check_paths
                    if x not in seen and not add_it_and_return_none(x)
                ]

            else:
                check_paths = [name_with_resolved_vars]

            for resource_path in check_paths:
                doc_uri = uris.from_fs_path(resource_path)
                resource_doc = ws.get_document(doc_uri, accept_from_file=True)
                if resource_doc is None:
                    continue
                return typing.cast(IRobotDocument, resource_doc)

            log.info(
                "Unable to find: %s (checked paths: %s)",
                name_with_resolved_vars,
                check_paths,
            )

        return None

    def get_variable_import_as_doc(
        self, variables_import: INode
    ) -> Optional[IRobotDocument]:
        return self.get_resource_import_as_doc(variables_import)

    @instance_cache
    def get_resource_imports_as_docs(
        self,
    ) -> Tuple[Tuple[IResourceImportNode, Optional[IRobotDocument]], ...]:
        ret: List[Tuple[IResourceImportNode, Optional[IRobotDocument]]] = []

        # Get keywords from resources
        resource_imports = self.get_resource_imports()
        for resource_import in resource_imports:
            resource_doc = self.get_resource_import_as_doc(resource_import)
            ret.append((resource_import, resource_doc))

        return tuple(ret)

    @instance_cache
    def get_resource_inits_as_docs(self) -> Tuple[IRobotDocument, ...]:
        doc = self.doc
        path = doc.path
        if not path:
            return ()

        if not os.path.isabs(path):
            # i.e.: We can't deal with untitled...
            return ()

        parent_dir_path = os.path.dirname(path)

        ret: List[IRobotDocument] = []
        ws = self.workspace
        folder_root_paths = ws.get_folder_paths()

        for root_path in folder_root_paths:
            if parent_dir_path.startswith(root_path):
                stop_condition = lambda parent_dir_path: parent_dir_path == root_path
                break

        else:
            # We're dealing with a resource out of the workspace root,
            # so, let's stop at max range.
            next_i: "partial[int]" = partial(next, itertools.count())
            stop_condition = lambda parent_dir_path: next_i() >= 6

        initial_parent_path = parent_dir_path

        def iter_inits():
            parent_dir_path = initial_parent_path
            while True:
                yield os.path.join(parent_dir_path, "__init__.robot")

                if stop_condition(parent_dir_path):
                    break

                initial_len = len(parent_dir_path)
                parent_dir_path = os.path.dirname(parent_dir_path)
                if initial_len == len(parent_dir_path) or not len(parent_dir_path):
                    break

        for resource_path in iter_inits():
            doc_uri = uris.from_fs_path(resource_path)
            resource_doc = ws.get_document(doc_uri, accept_from_file=True)
            if resource_doc is not None:
                ret.append(typing.cast(IRobotDocument, resource_doc))
        return tuple(ret)

    def iter_dependency_and_init_resource_docs(
        self, dependency_graph
    ) -> Iterator[IRobotDocument]:
        visited = set()
        for resource_doc in itertools.chain(
            (d[1] for d in dependency_graph.iter_all_resource_imports_with_docs()),
            iter(self.get_resource_inits_as_docs()),
        ):
            if resource_doc is not None:
                if resource_doc.uri not in visited:
                    visited.add(resource_doc.uri)
                    yield resource_doc

    @instance_cache
    def get_variable_imports_as_docs(
        self,
    ) -> Tuple[Tuple[IVariableImportNode, Optional[IRobotDocument]], ...]:
        ret: List[Tuple[IResourceImportNode, Optional[IRobotDocument]]] = []

        variable_imports = self.get_variable_imports()
        for variable_import in variable_imports:
            variable_doc = self.get_variable_import_as_doc(variable_import)
            ret.append((variable_import, variable_doc))

        return tuple(ret)

    @instance_cache
    def get_current_keyword_definition(self) -> Optional[IKeywordDefinition]:
        """
        Provides the current keyword even if we're in its arguments and not actually
        on the keyword itself.
        """
        current_keyword_definition_and_usage_info = (
            self.get_current_keyword_definition_and_usage_info()
        )
        if not current_keyword_definition_and_usage_info:
            return None
        return current_keyword_definition_and_usage_info[0]

    @instance_cache
    def get_current_keyword_usage_info(
        self,
    ) -> Optional[KeywordUsageInfo]:
        """
        Provides the current keyword even if we're in its arguments and not actually
        on the keyword itself.
        """
        from robotframework_ls.impl import ast_utils

        token_info = self.get_current_token()
        if token_info is None:
            return None
        cp: ICompletionContext = self

        while token_info.token.type == token_info.token.EOL:
            sel = cp.sel
            if sel.col > 0:
                cp = cp.create_copy_with_selection(sel.line, sel.col - 1)
                token_info = cp.get_current_token()
                if token_info is None:
                    return None
            else:
                break

        usage_info = ast_utils.create_keyword_usage_info_from_token(
            token_info.stack, token_info.node, token_info.token
        )
        return usage_info

    @instance_cache
    def get_current_keyword_definition_and_usage_info(
        self,
    ) -> Optional[Tuple[IKeywordDefinition, KeywordUsageInfo]]:
        """
        Provides the current keyword even if we're in its arguments and not actually
        on the keyword itself.
        """
        from robotframework_ls.impl.find_definition import find_keyword_definition

        usage_info = self.get_current_keyword_usage_info()
        if usage_info is not None:
            token = usage_info.token

            # token line is 1-based and col is 0-based (make both 0-based here).
            line = token.lineno - 1
            col = token.col_offset
            cp = self.create_copy_with_selection(line, col)
            definitions = find_keyword_definition(
                cp, TokenInfo(usage_info.stack, usage_info.node, usage_info.token)
            )
            if definitions:
                definition: IKeywordDefinition = next(iter(definitions))
                return definition, usage_info
        return None

    @instance_cache
    def collect_dependency_graph(self) -> ICompletionContextDependencyGraph:
        from robotframework_ls.impl.completion_context_dependency_graph import (
            CompletionContextDependencyGraph,
        )

        return CompletionContextDependencyGraph.from_completion_context(self)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: ICompletionContext = check_implements(self)



# Original work Copyright 2017 Palantir Technologies, Inc. (MIT)
# Original work Copyright 2020 Open Law Library. (Apache 2)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import os
from typing import Optional, Dict, List, Iterable, Tuple, Set

from robocorp_ls_core import uris
from robocorp_ls_core.basic import implements
from robocorp_ls_core.protocols import (
    IWorkspace,
    IDocument,
    IDocumentSelection,
    IWorkspaceFolder,
    IConfig,
)
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.uris import uri_scheme, to_fs_path, normalize_drive, normalize_uri
import threading
from robocorp_ls_core.lsp import (
    TextDocumentItem,
    TextDocumentContentChangeEvent,
    RangeTypedDict,
)
import weakref
from collections import namedtuple
import time
from robocorp_ls_core.watchdog_wrapper import IFSObserver

log = get_logger(__name__)

_FileMTimeInfo = namedtuple("_FileMTimeInfo", "st_mtime, st_size")


class _DirInfo(object):
    def __init__(self, scan_path):
        self.scan_path = scan_path
        self.files_in_directory: Set[str] = set()


class _VirtualFSThread(threading.Thread):

    SLEEP_AMONG_SCANS = 0.5
    INNER_SLEEP = 0.1

    def __init__(self, virtual_fs):
        from robocorp_ls_core.watchdog_wrapper import IFSWatch
        from robocorp_ls_core import load_ignored_dirs
        from robocorp_ls_core.callbacks import Callback

        threading.Thread.__init__(self)
        self.daemon = True

        self._virtual_fs = weakref.ref(virtual_fs)
        self.root_folder_path = virtual_fs.root_folder_path

        self.accept_directory = load_ignored_dirs.create_accept_directory_callable()
        self.accept_file = lambda path_name: path_name.endswith(
            tuple(virtual_fs._extensions)
        )
        self._disposed = threading.Event()
        self.first_check_done = threading.Event()
        self._check_done_events = []
        self._last_sleep = None
        self._fs_watch: Optional[IFSWatch] = None
        self._dirs_changed = set()
        self._trigger_loop = threading.Event()
        self.on_file_changed = Callback()

    def _check_need_sleep(self):
        last_sleep = self._last_sleep
        if last_sleep is None:
            self._last_sleep = time.time()
            return
        if time.time() - last_sleep > 0.3:
            time.sleep(self.INNER_SLEEP)
            self._last_sleep = time.time()

    def _check_dir(self, dir_path: str, directories: Set[str], level=0, recursive=True):
        # This is the actual poll loop
        if level > 20:  # At most 20 levels deep...
            log.critical(
                "Directory tree more than 20 levels deep: %s. Bailing out.", dir_path
            )
            return

        self._last_sleep = time.time()
        if self._disposed.is_set():
            return

        dir_path = normalize_drive(dir_path)
        directories.add(dir_path)
        dir_info = _DirInfo(dir_path)
        try:
            assert not isinstance(dir_path, bytes)
            if self._disposed.is_set():
                return

            i = 0
            for entry in os.scandir(dir_path):
                i += 1

                if i % 100 == 0:
                    self._check_need_sleep()
                if entry.is_dir():
                    if recursive and self.accept_directory(entry.path):
                        self._check_dir(entry.path, directories, level + 1)

                elif self.accept_file(entry.path):
                    dir_info.files_in_directory.add(normalize_drive(entry.path))

            virtual_fs = self._virtual_fs()
            if virtual_fs is None:
                return
            virtual_fs._dir_to_info[dir_path] = dir_info
        except OSError:
            pass  # Directory was removed in the meanwhile.

    def run(self):
        from robocorp_ls_core.watchdog_wrapper import PathInfo

        virtual_fs: _VirtualFS = self._virtual_fs()
        fs_observer: IFSObserver = virtual_fs._fs_observer

        # Setup tracking for changes
        self._fs_watch = fs_observer.notify_on_any_change(
            [PathInfo(self.root_folder_path, recursive=True)],
            on_change=self._on_change,
            extensions=virtual_fs._extensions,
        )
        check_done_events = self._check_done_events
        self._check_done_events = []

        # Do initial scan
        self._check_dir(self.root_folder_path, set())

        # Notify of initial scan
        self.first_check_done.set()
        self._notify_check_done_events(check_done_events)

        while not self._disposed.is_set():
            self._trigger_loop.wait(self.SLEEP_AMONG_SCANS)
            if self._disposed.is_set():
                return

            check_done_events = self._check_done_events
            self._check_done_events = []

            # Wait 100 more millis to catch up to further changes in the same loop.
            time.sleep(0.1)

            virtual_fs = self._virtual_fs()
            if virtual_fs is None:
                self.dispose()
                return

            # This would do a clean update, which'd be very cost intensive...
            # Instead, let's work only on the `_dirs_changed`.
            # directories = set()
            # current = list(virtual_fs._dir_to_info)
            # self._check_dir(self.root_folder_path, directories)
            #
            # for d in current:
            #     if d not in directories:
            #         virtual_fs._dir_to_info.pop(d, None)

            dirs_changed = self._dirs_changed
            self._dirs_changed = set()

            for dir_path in dirs_changed:
                dir_path = normalize_drive(dir_path)
                dir_info = _DirInfo(dir_path)
                try:
                    assert not isinstance(dir_path, bytes)
                    if self._disposed.is_set():
                        return

                    for entry in os.scandir(dir_path):
                        if not entry.is_dir() and self.accept_file(entry.path):
                            dir_info.files_in_directory.add(entry.path)
                except OSError:
                    if not os.path.exists(dir_path):
                        # Directory was removed.
                        virtual_fs._dir_to_info.pop(dir_path, None)
                else:
                    virtual_fs._dir_to_info[dir_path] = dir_info

            virtual_fs = None

            self._trigger_loop.clear()
            self._notify_check_done_events(check_done_events)

    def _on_change(self, src_path):
        changed_dir = os.path.dirname(src_path)
        self._dirs_changed.add(changed_dir)
        self._trigger_loop.set()
        self.on_file_changed(src_path)

    def dispose(self):
        fs_watch = self._fs_watch
        if fs_watch is not None:
            fs_watch.stop_tracking()
            self._fs_watch = None

        self._disposed.set()
        self._trigger_loop.set()

    def _notify_check_done_events(self, check_done_events):
        for event in check_done_events:
            event.set()

    def wait_for_check_done(self, timeout):
        event = threading.Event()
        self._check_done_events.append(event)
        if not event.wait(timeout):
            raise TimeoutError()


class _VirtualFS(object):
    def __init__(
        self, root_folder_path: str, extensions: Iterable[str], fs_observer: IFSObserver
    ):
        self.root_folder_path = normalize_drive(root_folder_path)

        self._dir_to_info: Dict[str, _DirInfo] = {}

        self._extensions = set(extensions)
        self._fs_observer = fs_observer

        # Do initial scan and then start tracking changes.
        self._virtual_fsthread = _VirtualFSThread(self)
        self._virtual_fsthread.start()
        self.on_file_changed = self._virtual_fsthread.on_file_changed

    def wait_for_check_done(self, timeout):
        self._virtual_fsthread.wait_for_check_done(timeout)

    def _iter_all_doc_uris(self, extensions: Tuple[str, ...]) -> Iterable[str]:
        """
        :param extensions:
            The extensions which are being searched (i.e.: ('.txt', '.py')).
        """
        assert self._extensions.issuperset(extensions)
        dir_infos = list(self._dir_to_info.values())
        for dir_info in dir_infos:
            for f in dir_info.files_in_directory:
                if f.endswith(extensions):
                    yield uris.from_fs_path(f)

    def dispose(self):
        self._virtual_fsthread.dispose()
        self._dir_to_info.clear()


class _WorkspaceFolderWithVirtualFS(object):
    """
    Walking a big tree may be time consuming, and very wasteful if users have
    things which the language server doesn't need (for instance, having a
    node_modules with thousands of unrelated files in the workspace).

    This class helps in keeping a cache just with the files we care about and
    invalidating them as needed.
    """

    def __init__(self, uri, name, track_file_extensions, fs_observer: IFSObserver):
        self.uri = uri
        self.name = name
        self.path = uris.to_fs_path(uri)

        self._vs: _VirtualFS = _VirtualFS(
            self.path, track_file_extensions, fs_observer=fs_observer
        )
        self.on_file_changed = self._vs.on_file_changed

    def _iter_all_doc_uris(self, extensions: Tuple[str, ...]) -> Iterable[str]:
        """
        :param extensions:
            The extensions which are being searched (i.e.: ('.txt', '.py')).
        """
        # Note: this function is meant to be thread-safe.

        vs = self._vs
        yield from vs._iter_all_doc_uris(extensions)

    def wait_for_check_done(self, timeout):
        self._vs.wait_for_check_done(timeout)

    def dispose(self):
        self._vs.dispose()

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IWorkspaceFolder = check_implements(self)


class Workspace(object):
    """
    Note: only a single thread can mutate the workspace, but multiple threads
    may read from it.
    """

    def __init__(
        self,
        root_uri: str,
        fs_observer: IFSObserver,
        workspace_folders: Optional[List[IWorkspaceFolder]] = None,
        track_file_extensions=(".robot", ".resource"),
    ) -> None:
        from robocorp_ls_core.lsp import WorkspaceFolder
        from robocorp_ls_core.callbacks import Callback

        self._main_thread = threading.current_thread()

        self._root_uri = root_uri
        self._root_uri_scheme = uri_scheme(self._root_uri)
        self._root_path = to_fs_path(self._root_uri)
        self._folders: Dict[str, _WorkspaceFolderWithVirtualFS] = {}
        self._track_file_extensions = track_file_extensions
        self._fs_observer = fs_observer

        # Contains the docs with files considered open.
        self._docs: Dict[str, IDocument] = {}

        # Contains the docs pointing to the filesystem.
        self._filesystem_docs: Dict[str, IDocument] = {}

        self.on_file_changed = Callback()

        if workspace_folders is not None:
            for folder in workspace_folders:
                self.add_folder(folder)

        if root_uri and root_uri not in self._folders:
            as_fs_path = uris.to_fs_path(root_uri)
            name = os.path.basename(as_fs_path)
            self.add_folder(WorkspaceFolder(root_uri, name))

    def on_changed_config(self, config: IConfig) -> None:
        pass

    def _check_in_mutate_thread(self):
        curr_thread = threading.current_thread()
        if self._main_thread is not curr_thread:
            raise AssertionError(
                f"Mutating the workspace can only be done at the thread: {self._main_thread}. Current thread: {curr_thread}"
            )

    def _create_document(
        self, doc_uri, source=None, version=None, force_load_source=False
    ):
        return Document(
            doc_uri,
            source=source,
            version=version,
            mutate_thread=self._main_thread,
            force_load_source=force_load_source,
        )

    def add_folder(self, folder: IWorkspaceFolder):
        """
        :param WorkspaceFolder folder:
        """
        self._check_in_mutate_thread()
        if folder.uri not in self._folders:
            folders = self._folders.copy()
            folder = _WorkspaceFolderWithVirtualFS(
                folder.uri,
                folder.name,
                track_file_extensions=self._track_file_extensions,
                fs_observer=self._fs_observer,
            )
            folder.on_file_changed.register(self.on_file_changed)
            folders[folder.uri] = folder
            self._folders = folders

    def remove_folder(self, folder_uri: str):
        self._check_in_mutate_thread()
        if folder_uri in self._folders:
            folders = self._folders.copy()
            folder = folders.pop(folder_uri)
            folder.on_file_changed.unregister(self.on_file_changed)
            folder.dispose()
            self._folders = folders

    @implements(IWorkspace.iter_documents)
    def iter_documents(self) -> Iterable[IDocument]:
        self._check_in_mutate_thread()  # i.e.: we don't really mutate here, but this is not thread safe.
        return self._docs.values()

    def get_open_docs_uris(self) -> List[str]:
        return list(d.uri for d in self._docs.values())

    @implements(IWorkspace.iter_folders)
    def iter_folders(self) -> Iterable[IWorkspaceFolder]:
        return (
            self._folders.values()
        )  # Ok, thread-safe (folders are always set as a whole)

    def wait_for_check_done(self, timeout):
        for folder in self.iter_folders():
            folder.wait_for_check_done(timeout)

    @implements(IWorkspace.get_folder_paths)
    def get_folder_paths(self) -> List[str]:
        folders = self._folders  # Ok, thread-safe (folders are always set as a whole)
        return [uris.to_fs_path(ws_folder.uri) for ws_folder in folders.values()]

    @implements(IWorkspace.get_document)
    def get_document(self, doc_uri: str, accept_from_file: bool) -> Optional[IDocument]:
        # Ok, thread-safe (does not mutate the _docs dict -- contents in the _filesystem_docs
        # may end up stale or we may have multiple loads when we wouldn't need,
        # but that should be ok).
        doc = self._docs.get(normalize_uri(doc_uri))
        if doc is not None:
            return doc

        if accept_from_file:
            doc = self._filesystem_docs.get(doc_uri)

            if doc is not None:
                if not doc.is_source_in_sync():
                    self._filesystem_docs.pop(doc_uri, None)
                    doc = None

            if doc is None:
                try:
                    doc = self._create_document(doc_uri, force_load_source=True)
                except:
                    log.debug("Unable to load contents from: %s", doc_uri)
                    # Unable to load contents: file does not exist.
                    doc = None
                else:
                    doc.immutable = True
                    self._filesystem_docs[doc_uri] = doc

        return doc

    def is_local(self):
        # Thread-safe (only accesses immutable data).
        return (
            self._root_uri_scheme == "" or self._root_uri_scheme == "file"
        ) and os.path.exists(self._root_path)

    @implements(IWorkspace.put_document)
    def put_document(self, text_document: TextDocumentItem) -> IDocument:
        self._check_in_mutate_thread()
        doc_uri = text_document.uri
        normalized_doc_uri = normalize_uri(doc_uri)
        doc = self._docs[normalized_doc_uri] = self._create_document(
            doc_uri, source=text_document.text, version=text_document.version
        )
        try:
            # In case the initial text wasn't passed, try to load it from source.
            # If it doesn't work, set the initial source as empty.
            # Note that we already checked the thread in this function, so, we
            # don't need to force it to be gotten in the constructor (which
            # could raise an exception).
            _source = doc.source
        except:
            doc.source = ""
        self._filesystem_docs.pop(normalized_doc_uri, None)
        return doc

    @implements(IWorkspace.remove_document)
    def remove_document(self, uri: str) -> None:
        self._check_in_mutate_thread()
        self._docs.pop(normalize_uri(uri), None)

    @property
    def root_path(self):
        # Thread-safe (only accesses immutable data).
        return self._root_path

    @property
    def root_uri(self) -> str:
        # Thread-safe (only accesses immutable data).
        return self._root_uri

    def update_document(
        self, text_doc: TextDocumentItem, change: TextDocumentContentChangeEvent
    ) -> IDocument:
        self._check_in_mutate_thread()
        doc_uri = text_doc["uri"]
        normalized_uri = normalize_uri(doc_uri)
        doc = self._docs[normalized_uri]

        # Note: don't mutate an existing doc, always create a new one based on it
        # (so, existing references won't have racing conditions).
        new_doc = self._create_document(doc_uri, doc.source, text_doc["version"])
        new_doc.apply_change(change)
        self._docs[normalized_uri] = new_doc
        return new_doc

    def iter_all_doc_uris_in_workspace(
        self, extensions: Tuple[str, ...]
    ) -> Iterable[str]:
        """
        :param extensions:
            The extensions which are being searched (i.e.: ('.txt', '.py')).
        """

        # Folders are set as a whole, so, this is thread safe.
        # This may be called in a thread.
        folders = self._folders.values()
        for folder in folders:
            yield from folder._iter_all_doc_uris(extensions)

    def dispose(self):
        self._check_in_mutate_thread()

        # Stop tracking folders and clear doc references.
        for folder_uri in list(self._folders.keys()):
            self.remove_folder(folder_uri)

        self._docs = {}
        self._filesystem_docs = {}

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IWorkspace = check_implements(self)


class Document(object):
    """
    Note: the doc isn't inherently thread-safe, so, the workspace should create
    a new document instead of mutating the source.

    Everything else (apart from changing the source) should be thread-safe
    (even without locks -- sometimes we may end up calculating things more than
    once, but that should not corrupt internal structures).
    """

    def __init__(
        self,
        uri: str,
        source=None,
        version: Optional[str] = None,
        *,
        mutate_thread=None,
        force_load_source=False,
    ):
        # During construction, set the mutate thread to the current thread.
        self._main_thread = threading.current_thread()
        self.immutable = False

        self.uri = uri
        self.version = version
        self.path = uris.to_fs_path(uri)  # Note: may be None.

        self._source = source
        self.__line_start_offsets = None

        # Only set when the source is read from disk.
        self._source_mtime = -1

        if force_load_source:
            # Just accessing should be ok to load the source.
            _ = self.source

        if mutate_thread is not None:
            # After construction it may only be mutated by the mutate thread if
            # specified.
            self._main_thread = mutate_thread

    def _check_in_mutate_thread(self):
        curr_thread = threading.current_thread()
        if self._main_thread is not curr_thread:
            raise AssertionError(
                f"Mutating the document can only be done at the thread: {self._main_thread}. Current thread: {curr_thread}"
            )

    def __str__(self):
        return str(self.uri)

    def __len__(self):
        return len(self.source)

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    def selection(self, line, col) -> IDocumentSelection:
        from robocorp_ls_core.document_selection import DocumentSelection

        return DocumentSelection(self, line, col)

    @property
    def _source(self) -> str:
        return self.__source

    @_source.setter
    def _source(self, source: str) -> None:
        # i.e.: when the source is set, reset the lines.
        self._check_in_mutate_thread()
        if self.immutable:
            raise RuntimeError(
                "This document is immutable, so, its source cannot be changed."
            )
        self.__source = source
        self._clear_caches()

    def _clear_caches(self):
        self._check_in_mutate_thread()
        self.__lines = None
        self.__line_start_offsets = None

    @property
    def _lines(self):
        lines = self.__lines
        if lines is None:
            lines = self.__lines = tuple(self.source.splitlines(True))
        return lines

    def get_internal_lines(self):
        return self._lines

    def iter_lines(self, keep_ends=True):
        lines = self._lines
        line = ""
        for line in lines:
            if keep_ends:
                yield line
            else:
                yield line.rstrip("\r\n")

        # If the last line ends with a new line, yield the final empty string.
        if line.endswith("\r") or line.endswith("\n"):
            yield ""

    def _compute_line_start_offsets(self):
        line_start_offset_to_info = self.__line_start_offsets
        if line_start_offset_to_info is None:

            line_start_offset_to_info = []
            offset = 0
            for line in self.iter_lines():
                line_start_offset_to_info.append(offset)
                offset += len(line)

        self.__line_start_offsets = line_start_offset_to_info
        return line_start_offset_to_info

    def offset_to_line_col(self, offset: int) -> Tuple[int, int]:
        if offset < 0:
            raise ValueError("Expected offset to be >0. Found: %s" % (offset,))

        import bisect

        line_start_offset_to_info = self._compute_line_start_offsets()
        i_line = bisect.bisect_left(line_start_offset_to_info, offset)
        if (
            i_line >= len(line_start_offset_to_info)
            or line_start_offset_to_info[i_line] > offset
        ):
            i_line -= 1
        line_start_offset = line_start_offset_to_info[i_line]
        return (i_line, offset - line_start_offset)

    def get_range(self, line: int, col: int, endline: int, endcol: int) -> str:
        if line >= self.get_line_count():
            return ""

        if line == endline:
            if endcol <= col:
                return ""

            line_contents = self.get_line(line)
            return line_contents[col:endcol]

        full_contents = []
        for i_line in range(line, min(endline + 1, self.get_line_count())):
            line_contents = self.get_line(i_line)
            if i_line == line:
                full_contents.append(self._lines[i_line][col:])
            elif i_line == endline:
                full_contents.append(self._lines[i_line][:endcol])
            else:
                full_contents.append(self._lines[i_line])
        return "".join(full_contents)

    def _load_source(self, mtime=None):
        self._check_in_mutate_thread()
        if mtime is None:
            mtime = os.path.getmtime(self.path)

        self._source_mtime = mtime
        with io.open(self.path, "r", encoding="utf-8") as f:
            self._source = f.read()

    @implements(IDocument.is_source_in_sync)
    def is_source_in_sync(self):
        try:
            mtime = os.path.getmtime(self.path)
            return self._source_mtime == mtime
        except Exception:
            log.info("Unable to get mtime for: %s", self.path)
            return False

    @property
    def source(self):
        if self._source is None:
            self._load_source()
        return self._source

    @source.setter
    def source(self, source):
        self._source = source

    @implements(IDocument.get_line)
    def get_line(self, line: int) -> str:
        try:
            return self._lines[line].rstrip("\r\n")
        except IndexError:
            return ""

    def get_last_line(self) -> str:
        try:
            last_line = self._lines[-1]
            if last_line.endswith("\r") or last_line.endswith("\n"):
                return ""
            return last_line
        except IndexError:
            return ""

    def get_last_line_col(self) -> Tuple[int, int]:
        lines = self._lines
        if not lines:
            return (0, 0)
        else:
            last_line = lines[-1]
            if last_line.endswith("\r") or last_line.endswith("\n"):
                return len(lines), 0
            return len(lines) - 1, len(last_line)

    def get_last_line_col_with_contents(self, contents: str) -> Tuple[int, int]:
        if not contents:
            raise ValueError("Contents not specified.")

        for i, line_contents in enumerate(self.iter_lines(keep_ends=False)):
            if contents in line_contents:
                return i, len(line_contents)
        raise RuntimeError(f"Unable to find line with contents: {contents}.")

    def get_line_count(self) -> int:
        lines = self._lines
        return len(lines)

    def apply_change(self, change: TextDocumentContentChangeEvent) -> None:
        """Apply a change to the document."""
        self._check_in_mutate_thread()
        text = change["text"]
        change_range: Optional[RangeTypedDict] = change.get("range")
        self._apply_change(change_range, text)

    def _apply_change(self, change_range: Optional[RangeTypedDict], text):
        self._check_in_mutate_thread()
        if not change_range:
            # The whole file has changed

            self._source = text
            return

        start_line = change_range["start"]["line"]
        start_col = change_range["start"]["character"]
        end_line = change_range["end"]["line"]
        end_col = change_range["end"]["character"]

        # Check for an edit occurring at the very end of the file
        if start_line == len(self._lines):

            self._source = self.source + text
            return

        new = io.StringIO()

        # Note: we could probably improve this sometime to work better with big documents
        # (but the link with the AST must be well thought out too).
        # References:
        # https://news.ycombinator.com/item?id=30415868
        # https://news.ycombinator.com/item?id=15381886
        # https://code.visualstudio.com/blogs/2018/03/23/text-buffer-reimplementation
        # https://blog.jetbrains.com/fleet/2022/02/fleet-below-deck-part-ii-breaking-down-the-editor/
        # https://raphlinus.github.io/xi/2020/06/27/xi-retrospective.html

        # Iterate over the existing document until we hit the edit range,
        # at which point we write the new text, then loop until we hit
        # the end of the range and continue writing.
        for i, line in enumerate(self._lines):
            if i < start_line:
                new.write(line)
                continue

            if i > end_line:
                new.write(line)
                continue

            if i == start_line:
                new.write(line[:start_col])
                new.write(text)

            if i == end_line:
                new.write(line[end_col:])

        self._source = new.getvalue()

    def apply_text_edits(self, text_edits):
        self._check_in_mutate_thread()
        for text_edit in reversed(text_edits):
            self._apply_change(text_edit["range"], text_edit["newText"])

    def find_line_with_contents(self, contents: str) -> int:
        """
        :param contents:
            The contents to be found.

        :return:
            The 0-based index of the contents.
        """
        for i, line in enumerate(self.iter_lines()):
            if contents in line:
                return i
        else:
            raise AssertionError(f"Did not find >>{contents}<< in doc.")

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IDocument = check_implements(self)



import time
start_time = time.time()

# do some busy work in parallel
print("Started main task")
x = 0
for i in range(1000000):
    x += 1
print("Completed main task")

if False:
    pass  # Breakpoint here

print('TotalTime>>%s<<' % (time.time() - start_time,))
print('TEST SUCEEDED')



import sys

from robocorp_ls_core.protocols import IConfig, IDirCache, IEndPoint

if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass


else:
    from typing import Protocol


# ===============================================================================
# EPConfigurationProvider
# ===============================================================================
class EPConfigurationProvider(Protocol):
    @property
    def config(self) -> IConfig:
        pass


class DefaultConfigurationProvider(object):
    def __init__(self, config: IConfig):
        self.config = config


# ===============================================================================
# EPDirCacheProvider
# ===============================================================================
class EPDirCacheProvider(Protocol):
    @property
    def dir_cache(self) -> IDirCache:
        pass


class DefaultDirCacheProvider(object):
    def __init__(self, dir_cache: IDirCache):
        self.dir_cache = dir_cache


# ===============================================================================
# EPEndPointProvider
# ===============================================================================
class EPEndPointProvider(Protocol):
    @property
    def endpoint(self) -> IEndPoint:
        pass


class DefaultEndPointProvider(object):
    def __init__(self, endpoint: IEndPoint):
        self.endpoint = endpoint



from robot.api.parsing import ModelTransformer, Token
from robot.utils.normalizing import normalize_whitespace

from robotidy.decorators import check_start_end_line


class NormalizeSettingName(ModelTransformer):
    """
    Normalize setting name.
    Ensure that setting names are title case without leading or trailing whitespace. For example from:

        *** Settings ***
        library    library.py
        test template    Template
        FORCE taGS    tag1

        *** Keywords ***
        Keyword
            [arguments]    ${arg}
            [ SETUP]   Setup Keyword

    To:

        *** Settings ***
        Library    library.py
        Test Template    Template
        Force Tags    tag1

        *** Keywords ***
        Keyword
            [Arguments]    ${arg}
            [Setup]   Setup Keyword

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/NormalizeSettingName.html for more examples.
    """

    @check_start_end_line
    def visit_Statement(self, node):  # noqa
        if node.type not in Token.SETTING_TOKENS:
            return node
        name = node.data_tokens[0].value
        if name.startswith("["):
            name = f"[{self.normalize_name(name[1:-1])}]"
        else:
            name = self.normalize_name(name)
        node.data_tokens[0].value = name
        return node

    @staticmethod
    def normalize_name(name):
        return normalize_whitespace(name).strip().title()



from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)


def get_robot_version() -> str:
    try:
        import robot  # noqa

        v = str(robot.get_version())
    except:
        log.exception("Unable to get robot version.")
        v = "unknown"
    return v


def get_robot_major_version() -> int:
    robot_version = get_robot_version()

    major_version = 4
    try:
        if "." in robot_version:
            major_version = int(robot_version.split(".")[0])
    except:
        log.exception("Unable to get robot major version.")

    return major_version



from robocorp_ls_core.protocols import IMonitor
from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)


class Monitor(object):
    def __init__(self, title: str = ""):
        self._title = title
        self._cancelled: bool = False

    def cancel(self) -> None:
        if self._title:
            log.info("Cancelled: %s", self._title)
        self._cancelled = True

    def check_cancelled(self) -> None:
        if self._cancelled:
            from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled

            raise JsonRpcRequestCancelled()

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: IMonitor = check_implements(self)



class T:

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return self.name


td = {T("foo", 24): "bar",
      T("gad", 42): "zooks",
      T("foo", 12): "bur"}

print('TEST SUCEEDED!')  # Break here



import time
import sys

try:
    try:
        from PySide import QtCore  # @UnresolvedImport
    except:
        from PySide2 import QtCore  # @UnresolvedImport
except:
    try:
        from PyQt4 import QtCore
    except:
        from PyQt5 import QtCore

# Subclassing QObject and using moveToThread
# http://labs.qt.nokia.com/2007/07/05/qthreads-no-longer-abstract/
class SomeObject(QtCore.QObject):

    try:
        finished = QtCore.Signal()  # @UndefinedVariable
    except:
        finished = QtCore.pyqtSignal()  # @UndefinedVariable

    def long_running(self):
        count = 0
        while count < 5:
            print("Increasing")  # break here
            count += 1
        self.finished.emit()

app = QtCore.QCoreApplication([])
objThread = QtCore.QThread()
obj = SomeObject()
obj.moveToThread(objThread)
obj.finished.connect(objThread.quit)
objThread.started.connect(obj.long_running)
objThread.finished.connect(app.exit)
objThread.start()
app.exec_()
print('TEST SUCEEDED!')



import pydevd
from _pydevd_bundle._debug_adapter import pydevd_schema

body = pydevd_schema.OutputEventBody('some output', 'my_category')
event = pydevd_schema.OutputEvent(body)
pydevd.send_json_message(event)

pydevd.send_json_message({
    "type": "event",
    "event": "output",
    "body": {"output": "some output 2", "category": "my_category2"}
})

print('TEST SUCEEDED')



from functools import lru_cache
from robocorp_ls_core.robotframework_log import get_logger
import re
from typing import Sequence

log = get_logger(__name__)


class TextUtilities(object):
    def __init__(self, text):
        self.text = text

    def strip_leading_chars(self, c):
        if len(c) != 1:
            raise AssertionError('Expected "%s" to have size == 1' % (c,))
        text = self.text
        if len(text) == 0 or text[0] != c:
            return False
        for i, text_c in enumerate(text):
            if text_c != c:
                self.text = self.text[i:]
                return True

        else:
            # Consumed all chars
            self.text = ""
            return True

    def strip(self):
        self.text = self.text.strip()


# Note: this not only makes it faster, but also makes us use less memory as a
# way to reuse the same 'interned" strings.
@lru_cache(maxsize=2000)
def normalize_robot_name(text: str) -> str:
    return text.lower().replace("_", "").replace(" ", "")


def is_variable_text(text: str) -> bool:
    from robotframework_ls.impl import robot_constants

    for p in robot_constants.VARIABLE_PREFIXES:
        if text.startswith(p + "{") and text.endswith("}"):
            return True
    return False


def contains_variable_text(text: str) -> bool:
    if "{" not in text:
        return False

    from robotframework_ls.impl import ast_utils

    token = ast_utils.create_token(text)

    try:
        tokenized_vars = ast_utils.tokenize_variables(token)
        for v in tokenized_vars:
            if v.type == v.VARIABLE:
                return True

    except:
        log.debug("Error tokenizing to variables: %s", text)

    return False


@lru_cache(600)
def matches_name_with_variables(name: str, name_with_variables: str) -> bool:
    """
    Checks if a given text matches a given keyword.

    Note: both should be already normalized.
    Note: should NOT be called if name_with_variables does not have '{' in it.

    :param name:
        The call that has resolved variables.
        i.e.: some call

    :param name_with_variables:
        The name which has variables.
        i.e.: some ${arg}
    """

    from robotframework_ls.impl import ast_utils

    try:
        tokenized_vars = ast_utils.tokenize_variables_from_name(name_with_variables)
    except:
        regexp = [re.escape(name_with_variables)]
    else:
        regexp = []
        for t in tokenized_vars:
            if t.type == t.VARIABLE:
                regexp.append("(.*)")
            else:
                regexp.append(re.escape(t.value))

    regexp.append("$")

    compiled = re.compile("".join(regexp))
    return bool(compiled.match(name))


def iter_dotted_names(text: str):
    """
    list(iter_dotted_names("")) == []
    list(iter_dotted_names("a")) == []
    list(iter_dotted_names("a.b")) == [("a", "b")]
    list(iter_dotted_names(".a.b.")) == [
        ("", "a.b."),
        (".a", "b."),
        (".a.b", ""),
    ]
    list(iter_dotted_names("a.b.")) == [("a", "b."), ("a.b", "")]
    list(iter_dotted_names("a.b.c")) == [("a", "b.c"), ("a.b", "c")]
    """
    splitted = text.split(".")
    splitted_len = len(splitted)
    if splitted_len > 1:
        import io

        buf = io.StringIO()

        for i, name in enumerate(splitted[:-1]):
            if i > 0:
                buf.write(".")
            buf.write(name)
            remainder = ".".join(splitted[i + 1 :])
            head = buf.getvalue()
            yield head, remainder


_DEPRECATED_PATTERN = re.compile(r"^\*DEPRECATED(.*)\*(.*)")


def has_deprecated_text(docs: str) -> bool:
    if docs and "DEPRECATED" in docs:
        matched = _DEPRECATED_PATTERN.match(docs)
        return bool(matched)

    return False


def build_keyword_docs_with_signature(
    keyword_name: str,
    args: Sequence[str],  # tuple(x.original_arg for x in keyword_args)
    docs: str,
    docs_format: str,
):

    if docs_format == "markdown":
        # Multi-line approach (it's a bit too big -- maybe as an option?)
        # if docs_format == "markdown":
        #     arg_docs = "  \n&nbsp;&nbsp;&nbsp;&nbsp;".join(
        #         ("**" + (x.replace("*", "\\*") + "**") for x in args)
        #     )
        #     return f"**{keyword_name}**  \n&nbsp;&nbsp;&nbsp;&nbsp;{arg_docs}\n\n{docs}"

        if args:
            escaped_args = (x.replace("*", "\\*") for x in args)
            arg_docs = f'({", ".join(escaped_args)})'

        else:
            arg_docs = ""
        return f"**{keyword_name}{arg_docs}**\n\n{docs}"
    else:
        if args:
            arg_docs = f'({", ".join(args)})'

        else:
            arg_docs = ""

        return f"{keyword_name}{arg_docs}\n\n{docs}"


@lru_cache(maxsize=300)
def get_digest_from_string(s: str) -> str:
    import hashlib

    return hashlib.sha256(s.encode("utf-8", "replace")).hexdigest()[:8]



import shutil
from pathlib import Path

from setuptools import find_packages, setup

_here = Path(__file__).parent.resolve()
_parent = _here.parent
_root = _parent.parent

_readme = _here / "README.md"
_thirdparty = _here / "ThirdPartyNotices.txt"
_license = _here / "LICENSE"
_copyright = _here / "COPYRIGHT"

# Note: always overwrite files from the original place if those exist.

_origin = _parent / _readme.name
if _origin.exists():
    shutil.copy2(_origin, _readme)

for path in [_thirdparty, _license, _copyright]:
    _origin = _root / path.name
    if _origin.exists():
        shutil.copy2(_origin, path)


def collect_pydevd_vendored_files():
    """Provides pydevd files relative to "robotframework_debug_adapter.vendored"."""
    import os

    VENDORED_ROOT = os.path.dirname(os.path.abspath(__file__))
    VENDORED_ROOT = os.path.join(
        VENDORED_ROOT, "robotframework_debug_adapter", "vendored"
    )
    ret = []
    for dir_, _, files in os.walk(VENDORED_ROOT):
        for file_name in files:
            if file_name.lower().endswith(
                (
                    ".py",
                    ".pyd",
                    ".so",
                    ".exe",
                    ".dll",
                    ".dylib",
                    ".bat",
                    ".c",
                    ".cpp",
                    ".h",
                    ".hpp",
                    ".sh",
                    ".pxd",
                    ".pyx",
                    ".md",
                    ".robot",
                    "makefile",
                    "license",
                )
            ):
                rel_dir = os.path.relpath(dir_, VENDORED_ROOT)
                rel_file = os.path.join(rel_dir, file_name)
                ret.append(rel_file)

    assert len(ret) > 50, "Did not collect vendored files properly. Found: %s" % (ret,)
    return ret


def collect_other_vendored_files():
    import os

    VENDORED_ROOT = os.path.dirname(os.path.abspath(__file__))
    VENDORED_ROOT = os.path.join(VENDORED_ROOT, "robotframework_ls", "vendored")
    ret = []
    for dir_, _, files in os.walk(VENDORED_ROOT):
        for file_name in files:
            if file_name.lower().endswith(
                (
                    ".py",
                    ".pyd",
                    ".so",
                    ".exe",
                    ".dll",
                    ".dylib",
                    ".bat",
                    ".c",
                    ".cpp",
                    ".h",
                    ".hpp",
                    ".sh",
                    ".pxd",
                    ".pyx",
                    ".md",
                    ".robot",
                    "makefile",
                    "license",
                )
            ):
                rel_dir = os.path.relpath(dir_, VENDORED_ROOT)
                rel_file = os.path.join(rel_dir, file_name)
                ret.append(rel_file)

    assert len(ret) > 20, "Did not collect vendored files properly. Found: %s" % (ret,)
    return ret


setup(
    name="robotframework-lsp",
    version="0.45.0",
    description="Language Server Protocol implementation for Robot Framework",
    long_description=_readme.read_text(),
    url="https://github.com/robocorp/robotframework-lsp",
    author="Fabio Zadrozny",
    license="Apache-2.0",
    copyright="Robocorp Technologies, Inc.",
    packages=find_packages(),
    zip_safe=False,
    long_description_content_type="text/markdown",
    python_requires=">=3.7",
    # List run-time dependencies here. These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["robotframework >=3.2"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[test]
    extras_require={
        "selenium": ["robotframework-seleniumlibrary >=4.4"],
        "test": [
            "mock",
            "pytest",
            "pytest-regressions==1.0.6",
            "pytest-xdist",
            "pytest-timeout",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Text Editors",
        "Topic :: Text Editors :: Integrated Development Environments (IDE)",
        "Topic :: Software Development :: Debuggers",
        "Framework :: Robot Framework",
        "Framework :: Robot Framework :: Tool",
    ],
    package_data={
        "robotframework_debug_adapter.vendored": collect_pydevd_vendored_files(),
        "robotframework_ls.vendored": collect_other_vendored_files(),
    },
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "console_scripts": ["robotframework_ls = robotframework_ls.__main__:main"],
        "jupyter_lsp_spec_v1": [
            "robotframework_ls = robotframework_ls.ext.jupyter_lsp:spec_v1"
        ],
    },
)



import sys
from _pydev_bundle._pydev_saved_modules import thread
import pycompletionserver
import socket
from urllib.parse import quote_plus

start_new_thread = thread.start_new_thread

BUILTIN_MOD = 'builtins'


def send(s, msg):
    s.send(bytearray(msg, 'utf-8'))


import unittest


class TestCPython(unittest.TestCase):

    def test_message(self):
        t = pycompletionserver.CompletionServer(0)

        l = []
        l.append(('Def', 'description'  , 'args'))
        l.append(('Def1', 'description1', 'args1'))
        l.append(('Def2', 'description2', 'args2'))

        msg = t.processor.format_completion_message(None, l)

        self.assertEqual('@@COMPLETIONS(None,(Def,description,args),(Def1,description1,args1),(Def2,description2,args2))END@@', msg)
        l = []
        l.append(('Def', 'desc,,r,,i()ption', ''))
        l.append(('Def(1', 'descriptio(n1', ''))
        l.append(('De,f)2', 'de,s,c,ription2', ''))
        msg = t.processor.format_completion_message(None, l)
        self.assertEqual('@@COMPLETIONS(None,(Def,desc%2C%2Cr%2C%2Ci%28%29ption, ),(Def%281,descriptio%28n1, ),(De%2Cf%292,de%2Cs%2Cc%2Cription2, ))END@@', msg)

    def create_connections(self):
        '''
        Creates the connections needed for testing.
        '''
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((pycompletionserver.HOST, 0))
        server.listen(1)  # socket to receive messages.

        t = pycompletionserver.CompletionServer(server.getsockname()[1])
        t.exit_process_on_kill = False
        start_new_thread(t.run, ())

        s, _addr = server.accept()

        return t, s

    def read_msg(self):
        finish = False
        msg = ''
        while finish == False:
            m = self.socket.recv(1024 * 4)
            m = m.decode('utf-8')
            if m.startswith('@@PROCESSING'):
                sys.stdout.write('Status msg: %s\n' % (msg,))
            else:
                msg += m

            if msg.find('END@@') != -1:
                finish = True

        return msg

    def test_completion_sockets_and_messages(self):
        t, socket = self.create_connections()
        self.socket = socket

        try:
            # now that we have the connections all set up, check the code completion messages.
            msg = quote_plus('math')
            send(socket, '@@IMPORTS:%sEND@@' % msg)  # math completions
            completions = self.read_msg()
            # print_ unquote_plus(completions)

            # math is a builtin and because of that, it starts with None as a file
            start = '@@COMPLETIONS(None,(__doc__,'
            start_2 = '@@COMPLETIONS(None,(__name__,'
            if ('/math.so,' in completions or
                '/math.cpython-33m.so,' in completions or
                '/math.cpython-34m.so,' in completions or
                'math.cpython-35m' in completions or
                'math.cpython-36m' in completions or
                'math.cpython-37m' in completions or
                'math.cpython-38' in completions or
                'math.cpython-39' in completions
                ):
                return
            self.assertTrue(completions.startswith(start) or completions.startswith(start_2), '%s DOESNT START WITH %s' % (completions, (start, start_2)))

            self.assertTrue('@@COMPLETIONS' in completions)
            self.assertTrue('END@@' in completions)

            # now, test i
            msg = quote_plus('%s.list' % BUILTIN_MOD)
            send(socket, "@@IMPORTS:%s\nEND@@" % msg)
            found = self.read_msg()
            self.assertTrue('sort' in found, 'Could not find sort in: %s' % (found,))

            # now, test search
            msg = quote_plus('inspect.ismodule')
            send(socket, '@@SEARCH%sEND@@' % msg)  # math completions
            found = self.read_msg()
            self.assertTrue('inspect.py' in found)
            for i in range(33, 100):
                if str(i) in found:
                    break
            else:
                self.fail('Could not find the ismodule line in %s' % (found,))

            # now, test search
            msg = quote_plus('inspect.CO_NEWLOCALS')
            send(socket, '@@SEARCH%sEND@@' % msg)  # math completions
            found = self.read_msg()
            self.assertTrue('inspect.py' in found)
            self.assertTrue('CO_NEWLOCALS' in found)

            # now, test search
            msg = quote_plus('inspect.BlockFinder.tokeneater')
            send(socket, '@@SEARCH%sEND@@' % msg)
            found = self.read_msg()
            self.assertTrue('inspect.py' in found)
#            self.assertTrue('CO_NEWLOCALS' in found)

        # reload modules test
#        send(socket, '@@RELOAD_MODULES_END@@')
#        ok = self.read_msg()
#        self.assertEqual('@@MSG_OK_END@@' , ok)
#        this test is not executed because it breaks our current enviroment.

        finally:
            try:
                sys.stdout.write('succedded...sending kill msg\n')
                self.send_kill_msg(socket)

#                while not hasattr(t, 'ended'):
#                    pass #wait until it receives the message and quits.

                socket.close()
                self.socket.close()
            except:
                pass

    def send_kill_msg(self, socket):
        socket.send(pycompletionserver.MSG_KILL_SERVER)




def get_here():
    a = 10


def foo(func):
    return func


def m1():  # @DontTrace
    get_here()


# @DontTrace
def m2():
    get_here()


# @DontTrace
@foo
def m3():
    get_here()


@foo
@foo
def m4():  # @DontTrace
    get_here()


def main():

    m1()  # break1

    m2()  # break2

    m3()  # break3

    m4()  # break4


if __name__ == '__main__':
    main()

    print('TEST SUCEEDED')



# WARNING: auto-generated file. Do NOT edit.
# If this file needs to be edited, change `codegen_kewords_in_args.py` and rerun.
KEYWORD_NAME_TO_KEYWORD_INDEX = {
    "runkeywords": 1,  # run_keywords
    "repeatkeyword": 2,  # repeat_keyword
    "runkeyword": 1,  # run_keyword
    "runkeywordandcontinueonfailure": 1,  # run_keyword_and_continue_on_failure
    "runkeywordandexpecterror": 2,  # run_keyword_and_expect_error
    "runkeywordandignoreerror": 1,  # run_keyword_and_ignore_error
    "runkeywordandreturn": 1,  # run_keyword_and_return
    "runkeywordandreturnif": 2,  # run_keyword_and_return_if
    "runkeywordandreturnstatus": 1,  # run_keyword_and_return_status
    "runkeywordandwarnonfailure": 1,  # run_keyword_and_warn_on_failure
    "runkeywordif": 2,  # run_keyword_if
    "runkeywordifalltestspassed": 1,  # run_keyword_if_all_tests_passed
    "runkeywordifanytestsfailed": 1,  # run_keyword_if_any_tests_failed
    "runkeywordiftestfailed": 1,  # run_keyword_if_test_failed
    "runkeywordiftestpassed": 1,  # run_keyword_if_test_passed
    "runkeywordiftimeoutoccurred": 1,  # run_keyword_if_timeout_occurred
    "runkeywordunless": 2,  # run_keyword_unless
    "waituntilkeywordsucceeds": 3,  # wait_until_keyword_succeeds
}
KEYWORD_NAME_TO_CONDITION_INDEX = {
    "continueforloopif": 1,  # continue_for_loop_if
    "exitforloopif": 1,  # exit_for_loop_if
    "passexecutionif": 1,  # pass_execution_if
    "returnfromkeywordif": 1,  # return_from_keyword_if
    "runkeywordandreturnif": 1,  # run_keyword_and_return_if
    "runkeywordif": 1,  # run_keyword_if
    "runkeywordunless": 1,  # run_keyword_unless
    "setvariableif": 1,  # set_variable_if
    "shouldbetrue": 1,  # should_be_true
    "shouldnotbetrue": 1,  # should_not_be_true
    "skipif": 1,  # skip_if
}



import unittest
import os
import sys
import pytest

# Note: ant.jar and junit.jar must be in the PYTHONPATH (see jython_test_deps)

IS_JYTHON = False
if sys.platform.find('java') != -1:
    IS_JYTHON = True
    from _pydev_bundle._pydev_jy_imports_tipper import ismethod
    from _pydev_bundle._pydev_jy_imports_tipper import isclass
    from _pydev_bundle._pydev_jy_imports_tipper import dir_obj
    from _pydev_bundle import _pydev_jy_imports_tipper
    from java.lang.reflect import Method  # @UnresolvedImport
    from java.lang import System  # @UnresolvedImport
    from java.lang import String  # @UnresolvedImport
    from java.lang.System import arraycopy  # @UnresolvedImport
    from java.lang.System import out  # @UnresolvedImport
    import java.lang.String  # @UnresolvedImport
    import org.python.core.PyDictionary  # @UnresolvedImport

__DBG = 0


def dbg(s):
    if __DBG:
        sys.stdout.write('%s\n' % (s,))


@pytest.mark.skipif(not IS_JYTHON, reason='Jython related test')
class TestMod(unittest.TestCase):

    def assert_args(self, tok, args, tips):
        for a in tips:
            if tok == a[0]:
                self.assertEqual(args, a[2])
                return
        raise AssertionError('%s not in %s', tok, tips)

    def assert_in(self, tok, tips):
        self.assertEqual(4, len(tips[0]))
        for a in tips:
            if tok == a[0]:
                return a
        s = ''
        for a in tips:
            s += str(a)
            s += '\n'
        raise AssertionError('%s not in %s' % (tok, s))

    def test_imports1a(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('java.util.HashMap')
        if f is None:
            return  # Not ok with java 9

        assert f.endswith('rt.jar')

    def test_imports1c(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('java.lang.Class')
        if f is None:
            return  # Not ok with java 9
        assert f.endswith('rt.jar')

    def test_imports1b(self):
        try:
            f, tip = _pydev_jy_imports_tipper.generate_tip('__builtin__.m')
            self.fail('err')
        except:
            pass

    def test_imports1(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('junit.framework.TestCase')
        assert f.endswith('junit.jar')
        ret = self.assert_in('assertEquals', tip)
#        self.assertEqual('', ret[2])

    def test_imports2(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('junit.framework')
        assert f.endswith('junit.jar')
        ret = self.assert_in('TestCase', tip)
        self.assertEqual('', ret[2])

    def test_imports2a(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('org.apache.tools.ant')
        assert f.endswith('ant.jar')
        ret = self.assert_in('Task', tip)
        self.assertEqual('', ret[2])

    def test_imports3(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('os')
        assert f.endswith('os.py')
        ret = self.assert_in('path', tip)
        self.assertEqual('', ret[2])

    def test_tip_on_string(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('string')
        self.assert_in('join', tip)
        self.assert_in('uppercase', tip)

    def test_imports(self):
        tip = _pydev_jy_imports_tipper.generate_tip('__builtin__')[1]
        self.assert_in('tuple'          , tip)
        self.assert_in('RuntimeError'   , tip)
        self.assert_in('RuntimeWarning' , tip)

    def test_imports5(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip('java.lang')
        if f is None:
            return  # Not ok with java 9
        assert f.endswith('rt.jar')
        tup = self.assert_in('String' , tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_CLASS), tup[3])

        tip = _pydev_jy_imports_tipper.generate_tip('java')[1]
        tup = self.assert_in('lang' , tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_IMPORT), tup[3])

        tip = _pydev_jy_imports_tipper.generate_tip('java.lang.String')[1]
        tup = self.assert_in('indexOf'          , tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])

        tip = _pydev_jy_imports_tipper.generate_tip('java.lang.String')[1]
        tup = self.assert_in('charAt'          , tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])
        self.assertEqual('(int)', tup[2])

        tup = self.assert_in('format'          , tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])
        self.assertEqual('(string, objectArray)', tup[2])
        self.assertTrue(tup[1].find('[Ljava.lang.Object;') == -1)

        tup = self.assert_in('getBytes', tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])
        assert '[B' not in tup[1]
        assert 'byte[]' in tup[1]

        f, tip = _pydev_jy_imports_tipper.generate_tip('__builtin__.str')
        assert f is None or f.endswith('jython.jar')  # Depends on jython version
        self.assert_in('find'          , tip)

        f, tip = _pydev_jy_imports_tipper.generate_tip('__builtin__.dict')
        assert f is None or f.endswith('jython.jar')  # Depends on jython version
        self.assert_in('get'          , tip)


@pytest.mark.skipif(not IS_JYTHON, reason='Jython related test')
class TestSearch(unittest.TestCase):

    def test_search_on_jython(self):
        assert _pydev_jy_imports_tipper.search_definition('os')[0][0].split(os.sep)[-1] in ('javaos.py', 'os.py')
        self.assertEqual(0, _pydev_jy_imports_tipper.search_definition('os')[0][1])

        assert _pydev_jy_imports_tipper.search_definition('os.makedirs')[0][0].split(os.sep)[-1] in ('javaos.py', 'os.py')
        self.assertNotEqual(0, _pydev_jy_imports_tipper.search_definition('os.makedirs')[0][1])

        # print _pydev_jy_imports_tipper.search_definition('os.makedirs')


@pytest.mark.skipif(not IS_JYTHON, reason='Jython related test')
class TestCompl(unittest.TestCase):

    def test_getting_info_on_jython(self):

        dbg('\n\n--------------------------- java')
        assert not ismethod(java)[0]
        assert not isclass(java)
        assert _pydev_jy_imports_tipper.ismodule(java)

        dbg('\n\n--------------------------- java.lang')
        assert not ismethod(java.lang)[0]
        assert not isclass(java.lang)
        assert _pydev_jy_imports_tipper.ismodule(java.lang)

        dbg('\n\n--------------------------- Method')
        assert not ismethod(Method)[0]
        assert isclass(Method)

        dbg('\n\n--------------------------- System')
        assert not ismethod(System)[0]
        assert isclass(System)

        dbg('\n\n--------------------------- String')
        assert not ismethod(System)[0]
        assert isclass(String)
        assert len(dir_obj(String)) > 10

        dbg('\n\n--------------------------- arraycopy')
        isMet = ismethod(arraycopy)
        assert isMet[0]
        assert isMet[1][0].basic_as_str() == "function:arraycopy args=['java.lang.Object', 'int', 'java.lang.Object', 'int', 'int'], varargs=None, kwargs=None, docs:None"
        assert not isclass(arraycopy)

        dbg('\n\n--------------------------- out')
        isMet = ismethod(out)
        assert not isMet[0]
        assert not isclass(out)

        dbg('\n\n--------------------------- out.println')
        isMet = ismethod(out.println)  # @UndefinedVariable
        assert isMet[0]
        assert len(isMet[1]) == 10
        self.assertEqual(isMet[1][0].basic_as_str(), "function:println args=[], varargs=None, kwargs=None, docs:None")
        assert isMet[1][1].basic_as_str() == "function:println args=['long'], varargs=None, kwargs=None, docs:None"
        assert not isclass(out.println)  # @UndefinedVariable

        dbg('\n\n--------------------------- str')
        isMet = ismethod(str)
        # the code below should work, but is failing on jython 22a1
        # assert isMet[0]
        # assert isMet[1][0].basic_as_str() == "function:str args=['org.python.core.PyObject'], varargs=None, kwargs=None, docs:None"
        assert not isclass(str)

        def met1():
            a = 3
            return a

        dbg('\n\n--------------------------- met1')
        isMet = ismethod(met1)
        assert isMet[0]
        assert isMet[1][0].basic_as_str() == "function:met1 args=[], varargs=None, kwargs=None, docs:None"
        assert not isclass(met1)

        def met2(arg1, arg2, *vararg, **kwarg):
            '''docmet2'''

            a = 1
            return a

        dbg('\n\n--------------------------- met2')
        isMet = ismethod(met2)
        assert isMet[0]
        assert isMet[1][0].basic_as_str() == "function:met2 args=['arg1', 'arg2'], varargs=vararg, kwargs=kwarg, docs:docmet2"
        assert not isclass(met2)

# Run for jython in command line:

# On Windows:
# c:/bin/jython2.7.0/bin/jython.exe -Dpython.path=jython_test_deps/ant.jar;jython_test_deps/junit.jar -m py.test tests/test_jysimpleTipper.py

# On Linux (different path separator for jars)
# jython -Dpython.path=jython_test_deps/ant.jar:jython_test_deps/junit.jar -m py.test tests/test_jysimpleTipper.py



import os
import sys
from collections import defaultdict
from difflib import unified_diff
from pathlib import Path
from typing import List, Tuple, Dict, Optional, Pattern

import click
from robot.api import get_model
from robot.errors import DataError

from robotidy.files import get_paths
from robotidy.transformers import load_transformers
from robotidy.utils import (
    StatementLinesCollector,
    decorate_diff_with_color,
    GlobalFormattingConfig,
    ModelWriter,
)


class Robotidy:
    def __init__(
        self,
        transformers: List[Tuple[str, List]],
        transformers_config: List[Tuple[str, List]],
        src: Tuple[str, ...],
        exclude: Optional[Pattern],
        extend_exclude: Optional[Pattern],
        overwrite: bool,
        show_diff: bool,
        formatting_config: GlobalFormattingConfig,
        verbose: bool,
        check: bool,
        output: Optional[Path],
        force_order: bool,
    ):
        self.sources = get_paths(src, exclude, extend_exclude)
        self.overwrite = overwrite
        self.show_diff = show_diff
        self.check = check
        self.verbose = verbose
        self.formatting_config = formatting_config
        self.output = output
        transformers_config = self.convert_configure(transformers_config)
        self.transformers = load_transformers(transformers, transformers_config, force_order=force_order)
        for transformer in self.transformers:
            # inject global settings TODO: handle it better
            setattr(transformer, "formatting_config", self.formatting_config)

    def transform_files(self):
        changed_files = 0
        for source in self.sources:
            try:
                stdin = False
                if str(source) == "-":
                    stdin = True
                    if self.verbose:
                        click.echo("Loading file from stdin")
                    source = self.load_from_stdin()
                elif self.verbose:
                    click.echo(f"Transforming {source} file")
                model = get_model(source)
                diff, old_model, new_model = self.transform(model)
                if diff:
                    changed_files += 1
                self.output_diff(model.source, old_model, new_model)
                if stdin:
                    self.print_to_stdout(new_model)
                elif diff:
                    self.save_model(model.source, model)
            except DataError:
                click.echo(
                    f"Failed to decode {source}. Default supported encoding by Robot Framework is UTF-8. Skipping file"
                )
                pass
        if not self.check or not changed_files:
            return 0
        return 1

    def transform(self, model):
        old_model = StatementLinesCollector(model)
        for transformer in self.transformers:
            transformer.visit(model)
        new_model = StatementLinesCollector(model)
        return new_model != old_model, old_model, new_model

    @staticmethod
    def load_from_stdin() -> str:
        return sys.stdin.read()

    def print_to_stdout(self, collected_lines):
        if not self.show_diff:
            click.echo(collected_lines.text)

    def save_model(self, source, model):
        if self.overwrite:
            output = self.output or model.source
            ModelWriter(output=output, newline=self.get_line_ending(source)).write(model)

    def get_line_ending(self, path: str):
        if self.formatting_config.line_sep == "auto":
            with open(path) as f:
                f.readline()
                if f.newlines is None:
                    return os.linesep
                if isinstance(f.newlines, str):
                    return f.newlines
                else:
                    return f.newlines[0]
        return self.formatting_config.line_sep

    def output_diff(
        self,
        path: str,
        old_model: StatementLinesCollector,
        new_model: StatementLinesCollector,
    ):
        if not self.show_diff:
            return
        old = [l + "\n" for l in old_model.text.splitlines()]
        new = [l + "\n" for l in new_model.text.splitlines()]
        lines = list(unified_diff(old, new, fromfile=f"{path}\tbefore", tofile=f"{path}\tafter"))
        if not lines:
            return
        colorized_output = decorate_diff_with_color(lines)
        click.echo(colorized_output.encode("ascii", "ignore").decode("ascii"), color=True)

    @staticmethod
    def convert_configure(configure: List[Tuple[str, List]]) -> Dict[str, List]:
        config_map = defaultdict(list)
        for transformer, args in configure:
            config_map[transformer].extend(args)
        return config_map



# Original work Copyright Fabio Zadrozny (EPL 1.0)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import json


from robocorp_ls_core.debug_adapter_core.dap import dap_base_schema as base_schema
from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import BaseSchema
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
    Breakpoint,
    ConfigurationDoneRequest,
    ConfigurationDoneResponse,
    ContinueRequest,
    DisconnectRequest,
    EvaluateRequest,
    InitializeRequest,
    InitializedEvent,
    LaunchRequest,
    LaunchResponse,
    NextRequest,
    PauseRequest,
    ScopesRequest,
    SetBreakpointsRequest,
    SetBreakpointsResponseBody,
    SetExceptionBreakpointsRequest,
    SourceBreakpoint,
    StackTraceRequest,
    StepInRequest,
    StepOutRequest,
    ThreadsRequest,
    VariablesRequest,
)
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_debug_adapter.constants import DEBUG
from typing import Optional


log = get_logger(__name__)


class DebugAdapterComm(object):
    """
    This is the class that actually processes commands from the client (VSCode)
    using the process stdin/stdout.

    It's responsible for actually starting the debugger process later on.

    It's created in the main thread and then control is passed on to the reader thread so that whenever
    something is read the json is handled by this processor.

    The queue it receives in the constructor should be used to talk to the writer thread, where it's expected
    to post protocol messages (which will be converted with 'to_dict()' and will have the 'seq' updated as
    needed).
    """

    def __init__(self, write_to_client_queue):
        from robotframework_debug_adapter.launch_process import LaunchProcess

        self.write_to_client_queue = write_to_client_queue
        self._launch_process: Optional[LaunchProcess] = None
        self._supports_run_in_terminal = False
        self._initialize_request_arguments = None
        self._run_in_debug_mode = True

        # i.e.: When a backend receives a stopped event, it sets itself as
        # a weakreference in weak_stopped_target_comm (so that we can redirect
        # related events, such as evaluate, stack trace, etc to the proper
        # backend).
        self.weak_stopped_target_comm = lambda: None

    @property
    def supports_run_in_terminal(self):
        return self._supports_run_in_terminal

    @property
    def initialize_request_arguments(self):
        return self._initialize_request_arguments

    def from_client(self, protocol_message):
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            READER_THREAD_STOPPED,
        )

        if protocol_message is READER_THREAD_STOPPED:
            if DEBUG:
                log.debug("%s: READER_THREAD_STOPPED." % (self.__class__.__name__,))
            return

        if DEBUG:
            log.debug(
                "Process json: %s\n"
                % (json.dumps(protocol_message.to_dict(), indent=4, sort_keys=True),)
            )

        if protocol_message.type == "request":
            method_name = "on_%s_request" % (protocol_message.command,)
            on_request = getattr(self, method_name, None)
            if on_request is not None:
                on_request(protocol_message)
            else:
                if DEBUG:
                    log.debug(
                        "Unhandled: %s not available in %s.\n"
                        % (method_name, self.__class__.__name__)
                    )

    def on_initialize_request(self, request: InitializeRequest):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            InitializeResponse,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Capabilities

        self._initialize_request_arguments = request.arguments
        initialize_response: InitializeResponse = build_response(request)
        self._supports_run_in_terminal = request.arguments.supportsRunInTerminalRequest
        capabilities: Capabilities = initialize_response.body
        capabilities.supportsConfigurationDoneRequest = True
        capabilities.supportsConditionalBreakpoints = True
        capabilities.supportsEvaluateForHovers = True
        capabilities.supportsHitConditionalBreakpoints = True
        capabilities.supportsLogPoints = True
        capabilities.exceptionBreakpointFilters = [
            {"filter": "logFailure", "label": "Robot Log FAIL", "default": True},
            {"filter": "logError", "label": "Robot Log ERROR", "default": True},
        ]
        # capabilities.supportsSetVariable = True
        self.write_to_client_message(initialize_response)

    def on_launch_request(self, request: LaunchRequest):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )
        from robotframework_debug_adapter.launch_process import LaunchProcess
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            SetDebuggerPropertyRequest,
            SetDebuggerPropertyArguments,
            AttachRequest,
            AttachRequestArguments,
        )

        self._run_in_debug_mode = run_in_debug_mode = not request.arguments.noDebug

        launch_response: LaunchResponse = build_response(request)
        launch_process = None
        try:

            self._launch_process = launch_process = LaunchProcess(
                request, launch_response, self
            )

            if launch_process.valid:
                # If on debug mode the launch is only considered finished when the connection
                # from the other side finishes properly.
                launch_process.launch()

                # Only write the initialized event after the process has been
                # launched so that we can forward breakpoints directly to the
                # target.
                self.write_to_client_message(InitializedEvent())
        except Exception as e:
            log.exception("Error launching.")
            launch_response.success = False
            launch_response.message = str(e)

        def on_finish_setup(*args, **kwargs):
            self.write_to_client_message(launch_response)  # acknowledge it
            if launch_process is not None:
                launch_process.after_launch_response_sent()

        if run_in_debug_mode and launch_process and launch_process.valid:
            # We must configure/attach to pydevd
            launch_process.write_to_pydevd(
                SetDebuggerPropertyRequest(
                    SetDebuggerPropertyArguments(
                        skipSuspendOnBreakpointException=("BaseException",),
                        skipPrintBreakpointException=("NameError",),
                        multiThreadsSingleNotification=True,
                    )
                )
            )

            attach_request_arguments = AttachRequestArguments(justMyCode=False)
            pydevd_attach_request = AttachRequest(attach_request_arguments)
            launch_process.write_to_pydevd(
                pydevd_attach_request, on_response=on_finish_setup
            )
        else:
            on_finish_setup()

    def on_configurationDone_request(self, request: ConfigurationDoneRequest):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        configuration_done_response: ConfigurationDoneResponse = build_response(request)
        launch_process = self._launch_process
        if launch_process is None:
            configuration_done_response.success = False
            configuration_done_response.message = (
                "Launch is not done (configurationDone uncomplete)."
            )
            self.write_to_client_message(configuration_done_response)
            return

        if launch_process.send_and_wait_for_configuration_done_request():
            self.write_to_client_message(configuration_done_response)  # acknowledge it

        else:
            # timed out
            configuration_done_response.success = False
            configuration_done_response.message = (
                "Timed out waiting for configurationDone event."
            )
            self.write_to_client_message(configuration_done_response)

    def on_disconnect_request(self, request: DisconnectRequest):
        disconnect_response = base_schema.build_response(request)

        if self._launch_process is not None:
            self._launch_process.disconnect(request)

        self.write_to_client_message(disconnect_response)

    def on_pause_request(self, request: PauseRequest):
        if self._run_in_debug_mode and self._launch_process is not None:
            self._launch_process.resend_request_to_pydevd(request)
        else:
            pause_response = base_schema.build_response(request)
            self.write_to_client_message(pause_response)

    def on_evaluate_request(self, request: EvaluateRequest):
        if self._launch_process is not None:
            if request.arguments.context == "repl":
                pass
                # i.e.: if not stopped anywhere we could send to the stdin...
                # self._launch_process.send_to_stdin(request.arguments.expression)
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_setExceptionBreakpoints_request(
        self, request: SetExceptionBreakpointsRequest
    ):
        if self._run_in_debug_mode and self._launch_process is not None:
            self._launch_process.resend_request_to_robot(request)
        else:
            response = base_schema.build_response(request)
            self.write_to_client_message(response)

    def on_setBreakpoints_request(self, request: SetBreakpointsRequest):
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            SetBreakpointsArguments,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Source

        if self._launch_process is None or not self._run_in_debug_mode:
            # Just acknowledge that no breakpoints are valid.
            breakpoints = []
            if request.arguments.breakpoints:
                for bp in request.arguments.breakpoints:
                    source_breakpoint = SourceBreakpoint(**bp)
                    breakpoints.append(
                        Breakpoint(
                            verified=False,
                            line=source_breakpoint.line,
                            source=request.arguments.source,
                        ).to_dict()
                    )

            self.write_to_client_message(
                base_schema.build_response(
                    request,
                    kwargs=dict(
                        body=SetBreakpointsResponseBody(breakpoints=breakpoints)
                    ),
                )
            )
            return

        arguments: SetBreakpointsArguments = request.arguments
        source: Source = arguments.source
        path = source.path
        if path and path.lower().endswith((".py", ".pyw")):
            self._launch_process.resend_request_to_pydevd(request)
        else:
            self._launch_process.resend_request_to_robot(request)

    def on_continue_request(self, request: ContinueRequest):
        if self._launch_process:
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_stepIn_request(self, request: StepInRequest):
        if self._launch_process:
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_stepOut_request(self, request: StepOutRequest):
        if self._launch_process:
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_stackTrace_request(self, request: StackTraceRequest):
        if self._launch_process:
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_next_request(self, request: NextRequest):
        if self._launch_process:
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_scopes_request(self, request: ScopesRequest):
        if self._launch_process:
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_variables_request(self, request: VariablesRequest):
        if self._launch_process:
            self._launch_process.resend_request_to_stopped_target(
                request, self.weak_stopped_target_comm()
            )

    def on_threads_request(self, request: ThreadsRequest):
        if self._launch_process:
            if self._run_in_debug_mode:
                self._launch_process.resend_request_to_pydevd(request)
            else:
                self._launch_process.resend_request_to_robot(request)

    def write_to_client_message(self, protocol_message: BaseSchema):
        """
        :param protocol_message:
            Some instance of one of the messages in the debug_adapter.schema.
        """
        self.write_to_client_queue.put(protocol_message)



import time
import sys

try:
    try:
        from PySide import QtCore  # @UnresolvedImport
    except:
        from PySide2 import QtCore  # @UnresolvedImport
except:
    try:
        from PyQt4 import QtCore
    except:
        from PyQt5 import QtCore

# Subclassing QThread
# http://doc.qt.nokia.com/latest/qthread.html
class AThread(QtCore.QThread):

    def run(self):
        count = 0
        while count < 5:
            print("Increasing", count)  # break here
            sys.stdout.flush()
            count += 1

app = QtCore.QCoreApplication([])
thread = AThread()
thread.finished.connect(app.exit)
thread.start()
app.exec_()
print('TEST SUCEEDED!')



# Based on: https://github.com/ESSS/pytest-regressions (License: MIT)
# Created copy because we need Python 2.6 which is not available on pytest-regressions.
# Note: only used for testing.

# encoding: UTF-8
import difflib
import pytest
import sys
from functools import partial

if sys.version_info[0] <= 2:
    from pathlib2 import Path
else:
    from pathlib import Path

FORCE_REGEN = False


@pytest.fixture
def original_datadir(request):
    # Method from: https://github.com/gabrielcnr/pytest-datadir
    # License: MIT
    import os.path
    return Path(os.path.splitext(request.module.__file__)[0])


@pytest.fixture
def datadir(original_datadir, tmpdir):
    # Method from: https://github.com/gabrielcnr/pytest-datadir
    # License: MIT
    import shutil
    result = Path(str(tmpdir.join(original_datadir.stem)))
    if original_datadir.is_dir():
        shutil.copytree(str(original_datadir), str(result))
    else:
        result.mkdir()
    return result


@pytest.fixture
def data_regression(datadir, original_datadir, request):
    return DataRegressionFixture(datadir, original_datadir, request)


def check_text_files(obtained_fn, expected_fn, fix_callback=lambda x: x, encoding=None):
    """
    Compare two files contents. If the files differ, show the diff and write a nice HTML
    diff file into the data directory.
    :param Path obtained_fn: path to obtained file during current testing.
    :param Path expected_fn: path to the expected file, obtained from previous testing.
    :param str encoding: encoding used to open the files.
    :param callable fix_callback:
        A callback to "fix" the contents of the obtained (first) file.
        This callback receives a list of strings (lines) and must also return a list of lines,
        changed as needed.
        The resulting lines will be used to compare with the contents of expected_fn.
    """
    __tracebackhide__ = True

    obtained_fn = Path(obtained_fn)
    expected_fn = Path(expected_fn)
    obtained_lines = fix_callback(obtained_fn.read_text(encoding=encoding).splitlines())
    expected_lines = expected_fn.read_text(encoding=encoding).splitlines()

    if obtained_lines != expected_lines:
        diff_lines = list(difflib.unified_diff(expected_lines, obtained_lines))
        if len(diff_lines) <= 500:
            html_fn = obtained_fn.with_suffix(".diff.html")
            try:
                differ = difflib.HtmlDiff()
                html_diff = differ.make_file(
                    fromlines=expected_lines,
                    fromdesc=expected_fn,
                    tolines=obtained_lines,
                    todesc=obtained_fn,
                )
            except Exception as e:
                html_fn = "(failed to generate html diff: %s)" % e
            else:
                html_fn.write_text(html_diff, encoding="UTF-8")

            diff = ["FILES DIFFER:", str(expected_fn), str(obtained_fn)]
            diff += ["HTML DIFF: %s" % html_fn]
            diff += diff_lines
            raise AssertionError("\n".join(diff))
        else:
            # difflib has exponential scaling and for thousands of lines it starts to take minutes to render
            # the HTML diff.
            msg = [
                "Files are different, but diff is too big (%s lines)" % (len(diff_lines),),
                "- obtained: %s" % (obtained_fn,),
                "- expected: %s" % (expected_fn,),
            ]
            raise AssertionError("\n".join(msg))


def perform_regression_check(
    datadir,
    original_datadir,
    request,
    check_fn,
    dump_fn,
    extension,
    basename=None,
    fullpath=None,
    obtained_filename=None,
    dump_aux_fn=lambda filename: [],
):
    """
    First run of this check will generate a expected file. Following attempts will always try to
    match obtained files with that expected file.
    :param Path datadir: Fixture embed_data.
    :param Path original_datadir: Fixture embed_data.
    :param SubRequest request: Pytest request object.
    :param callable check_fn: A function that receives as arguments, respectively, absolute path to
        obtained file and absolute path to expected file. It must assert if contents of file match.
        Function can safely assume that obtained file is already dumped and only care about
        comparison.
    :param callable dump_fn: A function that receive an absolute file path as argument. Implementor
        must dump file in this path.
    :param callable dump_aux_fn: A function that receives the same file path as ``dump_fn``, but may
        dump additional files to help diagnose this regression later (for example dumping image of
        3d views and plots to compare later). Must return the list of file names written (used to display).
    :param six.text_type extension: Extension of files compared by this check.
    :param six.text_type obtained_filename: complete path to use to write the obtained file. By
        default will prepend `.obtained` before the file extension.
    ..see: `data_regression.Check` for `basename` and `fullpath` arguments.
    """
    import re

    assert not (basename and fullpath), "pass either basename or fullpath, but not both"

    __tracebackhide__ = True

    if basename is None:
        basename = re.sub(r"[\W]", "_", request.node.name)

    if fullpath:
        filename = source_filename = Path(fullpath)
    else:
        filename = datadir / (basename + extension)
        source_filename = original_datadir / (basename + extension)

    def make_location_message(banner, filename, aux_files):
        msg = [banner, "- %s" % (filename,)]
        if aux_files:
            msg.append("Auxiliary:")
            msg += ["- %s" % (x,) for x in aux_files]
        return "\n".join(msg)

    if not filename.is_file():
        source_filename.parent.mkdir(parents=True, exist_ok=True)
        dump_fn(source_filename)
        aux_created = dump_aux_fn(source_filename)

        msg = make_location_message(
            "File not found in data directory, created:", source_filename, aux_created
        )
        pytest.fail(msg)
    else:
        if obtained_filename is None:
            if fullpath:
                obtained_filename = (datadir / basename).with_suffix(
                    ".obtained" + extension
                )
            else:
                obtained_filename = filename.with_suffix(".obtained" + extension)

        dump_fn(obtained_filename)

        try:
            check_fn(obtained_filename, filename)
        except AssertionError:
            if FORCE_REGEN:
                dump_fn(source_filename)
                aux_created = dump_aux_fn(source_filename)
                msg = make_location_message(
                    "Files differ and FORCE_REGEN set, regenerating file at:",
                    source_filename,
                    aux_created,
                )
                pytest.fail(msg)
            else:
                dump_aux_fn(obtained_filename)
                raise


class DataRegressionFixture(object):
    """
    Implementation of `data_regression` fixture.
    """

    def __init__(self, datadir, original_datadir, request):
        """
        :type datadir: Path
        :type original_datadir: Path
        :type request: FixtureRequest
        """
        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir

    def check(self, data_dict, basename=None, fullpath=None):
        """
        Checks the given dict against a previously recorded version, or generate a new file.
        :param dict data_dict: any yaml serializable dict.
        :param str basename: basename of the file to test/record. If not given the name
            of the test is used.
            Use either `basename` or `fullpath`.
        :param str fullpath: complete path to use as a reference file. This option
            will ignore ``datadir`` fixture when reading *expected* files but will still use it to
            write *obtained* files. Useful if a reference file is located in the session data dir for example.
        ``basename`` and ``fullpath`` are exclusive.
        """
        __tracebackhide__ = True

        def dump(filename):
            """Dump dict contents to the given filename"""
            import json

            s = json.dumps(data_dict, sort_keys=True, indent=4)
            if isinstance(s, bytes):
                s = s.decode('utf-8')

            s = u'\n'.join([line.rstrip() for line in s.splitlines()])
            s = s.encode('utf-8')

            with filename.open("wb") as f:
                f.write(s)

        perform_regression_check(
            datadir=self.datadir,
            original_datadir=self.original_datadir,
            request=self.request,
            check_fn=partial(check_text_files, encoding="UTF-8"),
            dump_fn=dump,
            extension=".json",
            basename=basename,
            fullpath=fullpath,
        )
